# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

from itertools import chain
from json import dumps, loads
from zipfile import ZIP_DEFLATED, ZipFile

from pytest import mark, raises

from ..target import AssetManager
from .storage import TestCase, TestCaseLoadFailure, TestFileExists


def test_testcase_01(tmp_path):
    """test empty TestCase"""
    l_page = "land.html"
    r_page = "redirect.html"
    adpt_name = "test-adapter"
    with TestCase(l_page, r_page, adpt_name) as tcase:
        assert tcase.landing_page == l_page
        assert tcase.redirect_page == r_page
        assert tcase.assets is None
        assert tcase.adapter_name == adpt_name
        assert tcase.duration is None
        assert tcase.data_size == 0
        assert tcase.input_fname is None
        assert tcase.timestamp > 0
        assert not tcase.env_vars
        assert tcase.data_path
        assert not tcase._files.optional
        assert not tcase._files.required
        assert not any(tcase.contents)
        assert tcase.pop_assets() is None
        assert not any(tcase.optional)
        tcase.dump(tmp_path)
        assert not any(tmp_path.iterdir())
        tcase.dump(tmp_path, include_details=True)
        assert (tmp_path / "test_info.json").is_file()


@mark.parametrize(
    "copy, required",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_testcase_02(tmp_path, copy, required):
    """test TestCase.add_from_file()"""
    with TestCase("land_page.html", "a.html", "adpt", input_fname="in.bin") as tcase:
        in_file = tmp_path / "file.bin"
        in_file.write_text("data")
        tcase.add_from_file(in_file, copy=copy, required=required)
        assert tcase.data_size == 4
        assert "file.bin" in tcase.contents
        if required:
            assert in_file.name not in tcase.optional
        else:
            assert in_file.name in tcase.optional
        assert in_file.exists() == copy
        # try overwriting existing file
        if copy:
            with raises(TestFileExists, match="'file.bin' exists in test"):
                tcase.add_from_file(in_file, copy=True)
            assert in_file.exists()
        else:
            assert not in_file.exists()


@mark.parametrize(
    "file_paths",
    [
        ("a.bin",),
        ("a/a.bin",),
        ("a.bin", "b.bin"),
        ("a.bin", "b/c.bin", "b/d.bin"),
    ],
)
def test_testcase_03(tmp_path, file_paths):
    """test TestCase.add_from_file()"""
    with TestCase("land_page.html", "a.html", "adpt") as tcase:
        for file_path in file_paths:
            src_file = tmp_path / file_path
            src_file.parent.mkdir(exist_ok=True, parents=True)
            src_file.write_text("data")
            tcase.add_from_file(src_file, file_name=file_path, required=True)
            assert file_path in tcase.contents
            assert file_path not in tcase.optional


def test_testcase_04():
    """test TestCase.add_from_bytes()"""
    with TestCase("a.html", None, "adpt") as tcase:
        tcase.add_from_bytes(b"foo", "a.html", required=True)
        tcase.add_from_bytes(b"foo", "b.html", required=False)
        assert "a.html" in (x.file_name for x in tcase._files.required)
        assert "b.html" in (x.file_name for x in tcase._files.optional)
        # add file with invalid file name
        with raises(ValueError, match="invalid path ''"):
            tcase.add_from_bytes(b"foo", "", required=False)


def test_testcase_05():
    """test TestCase.purge_optional()"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        # no optional files
        tcase.purge_optional(["foo"])
        # setup
        tcase.add_from_bytes(b"foo", "testfile1.bin", required=True)
        tcase.add_from_bytes(b"foo", "testfile2.bin", required=False)
        tcase.add_from_bytes(b"foo", "testfile3.bin", required=False)
        tcase.add_from_bytes(b"foo", "not_served.bin", required=False)
        assert len(tuple(tcase.optional)) == 3
        # nothing to remove - with required
        tcase.purge_optional(chain(["testfile1.bin"], tcase.optional))
        assert len(tuple(tcase.optional)) == 3
        # nothing to remove - use relative path (forced)
        tcase.purge_optional(x.file_name for x in tcase._files.optional)
        assert len(tuple(tcase.optional)) == 3
        # nothing to remove - use absolute path
        tcase.purge_optional(x.data_file.as_posix() for x in tcase._files.optional)
        assert len(tuple(tcase.optional)) == 3
        # remove not_served.bin
        tcase.purge_optional(["testfile2.bin", "testfile3.bin"])
        assert len(tuple(tcase.optional)) == 2
        assert "testfile2.bin" in tcase.optional
        assert "testfile3.bin" in tcase.optional
        assert "not_served.bin" not in tcase.optional
        # remove remaining optional
        tcase.purge_optional(["testfile1.bin"])
        assert not any(tcase.optional)


def test_testcase_06():
    """test TestCase.data_size"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        assert tcase.data_size == 0
        tcase.add_from_bytes(b"1", "testfile1.bin", required=True)
        assert tcase.data_size == 1
        tcase.add_from_bytes(b"12", "testfile2.bin", required=False)
        assert tcase.data_size == 3


def test_testcase_07(tmp_path):
    """test TestCase.load_single() using a directory - fail cases"""
    # missing test_info.json
    with raises(TestCaseLoadFailure, match="Missing 'test_info.json'"):
        TestCase.load_single(tmp_path)
    # invalid test_info.json
    (tmp_path / "test_info.json").write_bytes(b"X")
    with raises(TestCaseLoadFailure, match="Invalid 'test_info.json'"):
        TestCase.load_single(tmp_path)
    # test_info.json missing 'target' entry
    (tmp_path / "test_info.json").write_bytes(b"{}")
    with raises(
        TestCaseLoadFailure, match="'test_info.json' has invalid 'target' entry"
    ):
        TestCase.load_single(tmp_path)
    # build a test case
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_from_file(entry_point)
        src.dump(src_dir, include_details=True)
    # bad 'target' entry in test_info.json
    entry_point.unlink()
    with raises(TestCaseLoadFailure, match="Entry point 'target.bin' not found in"):
        TestCase.load_single(src_dir)
    # bad 'env' entry in test_info.json
    entry_point.touch()
    with AssetManager(base_path=str(tmp_path)) as assets:
        (tmp_path / "example_asset").touch()
        assets.add("example", str(tmp_path / "example_asset"), copy=False)
        with TestCase("target.bin", None, "test-adapter") as src:
            src.assets = assets
            src.dump(src_dir, include_details=True)
    test_info = loads((src_dir / "test_info.json").read_text())
    test_info["env"] = {"bad": 1}
    (src_dir / "test_info.json").write_text(dumps(test_info))
    with raises(TestCaseLoadFailure, match="'env' contains invalid entries"):
        TestCase.load_single(src_dir)


def test_testcase_08(mocker, tmp_path):
    """test TestCase.load_single() using a directory"""
    # build a valid test case
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    asset_file = src_dir / "example_asset"
    asset_file.touch()
    (src_dir / "optional.bin").touch()
    (src_dir / "x.bin").touch()
    nested = src_dir / "nested"
    nested.mkdir()
    # overlap file name in different directories
    (nested / "x.bin").touch()
    (tmp_path / "src" / "nested" / "empty").mkdir()
    dst_dir = tmp_path / "dst"
    with AssetManager(base_path=str(tmp_path)) as assets:
        assets.add("example", str(asset_file))
        with TestCase("target.bin", None, "test-adapter") as src:
            src.env_vars["TEST_ENV_VAR"] = "100"
            src.add_from_file(entry_point)
            src.add_from_file(src_dir / "optional.bin", required=False)
            src.add_from_file(src_dir / "x.bin", required=False)
            src.add_from_file(
                nested / "x.bin",
                file_name=str((nested / "x.bin").relative_to(src_dir)),
                required=False,
            )
            src.assets = assets
            src.dump(dst_dir, include_details=True)
    # test loading test case from test_info.json
    with TestCase.load_single(dst_dir) as dst:
        asset = dst.pop_assets()
        assert asset
        with asset:
            assert "example" in asset.assets
        assert dst.landing_page == "target.bin"
        assert "target.bin" in (x.file_name for x in dst._files.required)
        assert "optional.bin" in (x.file_name for x in dst._files.optional)
        assert "x.bin" in (x.file_name for x in dst._files.optional)
        assert "nested/x.bin" in (x.file_name for x in dst._files.optional)
        assert dst.env_vars["TEST_ENV_VAR"] == "100"
        assert dst.timestamp > 0
    # test load with missing asset
    mocker.patch("grizzly.common.storage.AssetManager.load", side_effect=OSError)
    with raises(TestCaseLoadFailure):
        TestCase.load_single(dst_dir)


def test_testcase_09(tmp_path):
    """test TestCase.load_single() using a file"""
    # invalid entry_point specified
    with raises(TestCaseLoadFailure, match="Missing or invalid TestCase"):
        TestCase.load_single(tmp_path / "missing_file", adjacent=False)
    # valid test case
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    (src_dir / "optional.bin").touch()
    # load single file test case
    with TestCase.load_single(entry_point, adjacent=False) as tcase:
        assert tcase.assets is None
        assert not tcase.env_vars
        assert tcase.landing_page == "target.bin"
        assert "target.bin" in (x.file_name for x in tcase._files.required)
        assert "optional.bin" not in (x.file_name for x in tcase._files.optional)
        assert tcase.timestamp == 0
    # load full test case
    with TestCase.load_single(entry_point, adjacent=True) as tcase:
        assert tcase.landing_page == "target.bin"
        assert "target.bin" in (x.file_name for x in tcase._files.required)
        assert "optional.bin" in (x.file_name for x in tcase._files.optional)


def test_testcase_10(tmp_path):
    """test TestCase - dump, load and compare"""
    working = tmp_path / "working"
    working.mkdir()
    with TestCase("a.html", "b.html", "adpt") as org:
        # set non default values
        org.duration = 1.23
        org.env_vars = {"en1": "1", "en2": "2"}
        org.hang = True
        org.input_fname = "infile"
        org.time_limit = 10
        org.add_from_bytes(b"a", "a.html")
        with AssetManager(base_path=str(tmp_path)) as assets:
            fake = tmp_path / "fake_asset"
            fake.touch()
            assets.add("fake", str(fake))
            org.assets = assets
            org.dump(working, include_details=True)
        org.assets = None
        with TestCase.load_single(working, adjacent=False) as loaded:
            try:
                for prop in TestCase.__slots__:
                    if prop.startswith("_") or prop in ("assets", "redirect_page"):
                        continue
                    assert getattr(loaded, prop) == getattr(org, prop)
                assert not set(org.contents) ^ set(loaded.contents)
                assert loaded.assets
                assert "fake" in loaded.assets.assets
            finally:
                if loaded.assets:
                    loaded.assets.cleanup()


def test_testcase_11(tmp_path):
    """test TestCase.load() - missing file and empty directory"""
    # missing file
    with raises(TestCaseLoadFailure, match="Invalid TestCase path"):
        TestCase.load(tmp_path / "missing")
    # empty path
    assert not TestCase.load(tmp_path, adjacent=True)


def test_testcase_12(tmp_path):
    """test TestCase.load() - single file"""
    tfile = tmp_path / "testcase.html"
    tfile.touch()
    testcases = TestCase.load(tfile, adjacent=False)
    try:
        assert len(testcases) == 1
        assert all(x.assets is None for x in testcases)
    finally:
        any(x.cleanup() for x in testcases)


def test_testcase_13(tmp_path):
    """test TestCase.load() - single directory"""
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_from_bytes(b"test", "target.bin")
        src.dump(tmp_path, include_details=True)
    testcases = TestCase.load(tmp_path)
    try:
        assert len(testcases) == 1
        assert all(x.assets is None for x in testcases)
    finally:
        any(x.cleanup() for x in testcases)


def test_testcase_14(tmp_path):
    """test TestCase.load() - multiple directories (with assets)"""
    nested = tmp_path / "nested"
    nested.mkdir()
    asset_file = tmp_path / "example_asset"
    asset_file.touch()
    with AssetManager(base_path=str(tmp_path)) as assets:
        assets.add("example", str(asset_file))
        with TestCase("target.bin", None, "test-adapter") as src:
            src.assets = assets
            src.add_from_bytes(b"test", "target.bin")
            src.dump(nested / "test-1", include_details=True)
            src.dump(nested / "test-2", include_details=True)
            src.dump(nested / "test-3", include_details=True)
    testcases = TestCase.load(nested)
    try:
        assert len(testcases) == 3
        assert all(x.assets is not None for x in testcases)
        asset = testcases[-1].pop_assets()
        assert asset is not None
        assert "example" in asset.assets
    finally:
        any(x.cleanup() for x in testcases)
        for test in testcases:
            if test.assets:
                test.assets.cleanup()
    # try loading testcases that are nested too deep
    assert not TestCase.load(tmp_path)


def test_testcase_15(tmp_path):
    """test TestCase.load() - archive"""
    archive = tmp_path / "testcase.zip"
    # bad archive
    archive.write_bytes(b"x")
    with raises(TestCaseLoadFailure, match="Testcase archive is corrupted"):
        TestCase.load(archive)
    # build archive containing multiple testcases
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_from_bytes(b"test", "target.bin")
        src.dump(tmp_path / "test-0", include_details=True)
        src.dump(tmp_path / "test-1", include_details=True)
        src.dump(tmp_path / "test-2", include_details=True)
    (tmp_path / "log_dummy.txt").touch()
    (tmp_path / "not_a_tc").mkdir()
    (tmp_path / "not_a_tc" / "file.txt").touch()
    with ZipFile(archive, mode="w", compression=ZIP_DEFLATED) as zfp:
        for entry in tmp_path.rglob("*"):
            if entry.is_file():
                zfp.write(str(entry), arcname=str(entry.relative_to(tmp_path)))
    testcases = TestCase.load(archive)
    try:
        assert len(testcases) == 3
        assert all(x.assets is None for x in testcases)
        assert all("target.bin" in x.contents for x in testcases)
    finally:
        any(x.cleanup() for x in testcases)


def test_testcase_16(tmp_path):
    """test TestCase.add_batch()"""
    include = tmp_path / "inc_path"
    include.mkdir()
    inc_1 = include / "file.bin"
    inc_1.write_bytes(b"a")
    (include / "nested").mkdir()
    inc_2 = include / "nested" / "nested.js"
    inc_2.write_bytes(b"a")
    other_path = tmp_path / "other_path"
    other_path.mkdir()
    (other_path / "no_include.bin").write_bytes(b"a")
    with TestCase("a.b", "a.b", "simple") as tcase:
        # missing directory
        tcase.add_batch("/missing/path/", tuple())
        assert not any(tcase.contents)
        # missing file
        with raises(IOError):
            tcase.add_batch(tmp_path, [tmp_path / "missing.bin"])
        assert not any(tcase.contents)
        # relative file name
        tcase.add_batch(include, ["file.bin"])
        assert not any(tcase.contents)
        # valid list
        tcase.add_batch(include, [inc_1, inc_2, tmp_path / "inc_path2" / "extra.bin"])
        assert "file.bin" in tcase.contents
        assert "nested/nested.js" in tcase.contents
        assert len(list(tcase.contents)) == 2
        # nested url
        tcase.add_batch(include, [inc_1], prefix="test")
        assert "test/file.bin" in tcase.contents
        assert len(list(tcase.contents)) == 3
        # collision
        with raises(TestFileExists, match="'file.bin' exists in test"):
            tcase.add_batch(include, [inc_1])


def test_testcase_17(tmp_path):
    """test TestCase.scan_path()"""
    # empty path
    (tmp_path / "not-test").mkdir()
    assert not any(TestCase.scan_path(tmp_path))
    # multiple test case directories
    paths = [tmp_path / f"test-{i}" for i in range(3)]
    with TestCase("test.htm", None, "test-adapter") as src:
        src.add_from_bytes(b"test", "test.htm")
        for path in paths:
            src.dump(path, include_details=True)
    tc_paths = list(TestCase.scan_path(tmp_path))
    assert len(tc_paths) == 3
    # single test case directory
    tc_paths = list(TestCase.scan_path(paths[0]))
    assert len(tc_paths) == 1


def test_testcase_18():
    """test TestCase.get_file()"""
    with TestCase("test.htm", None, "test-adapter") as src:
        src.add_from_bytes(b"test", "test.htm")
        assert src.get_file("missing") is None
        assert src.get_file("test.htm")


def test_testcase_19():
    """test TestCase.clone()"""
    with TestCase("a.htm", "b.htm", "adpt", input_fname="fn", time_limit=2) as src:
        src.duration = 1.2
        src.hang = True
        src.add_from_bytes(b"123", "test.htm", required=True)
        src.add_from_bytes(b"456", "opt.htm", required=False)
        src.env_vars["foo"] = "bar"
        with src.clone() as dst:
            for prop in TestCase.__slots__:
                if prop.startswith("_"):
                    continue
                assert getattr(src, prop) == getattr(dst, prop)
            assert src.data_size == dst.data_size
            for file, data in (
                ("test.htm", b"123"),
                ("opt.htm", b"456"),
            ):
                src_file = src.get_file(file).data_file
                dst_file = dst.get_file(file).data_file
                assert src_file.read_bytes() == data
                assert dst_file.read_bytes() == data
                assert not dst_file.samefile(src_file)
            assert dst.env_vars == {"foo": "bar"}
            assert not set(src.optional) ^ set(dst.optional)


@mark.parametrize(
    "path",
    [
        # empty file name
        "",
        # path (root) with missing file name
        "/",
        # path (root) with missing file name
        "/.",
        # path with missing file name
        "path/",
        # outside of wwwroot
        "../test",
        # outside of wwwroot
        "a/../../b",
        # outside of wwwroot
        "..",
        # outside of wwwroot
        "C:\\test",
        # cwd
        ".",
    ],
)
def test_testcase_20(path):
    """test TestCase.sanitize_path() with invalid paths"""
    with raises(ValueError, match="invalid path"):
        TestCase.sanitize_path(path)


@mark.parametrize(
    "path, expected_result",
    [
        ("a", "a"),
        ("file.bin", "file.bin"),
        ("a/file.bin", "a/file.bin"),
        ("/file.bin", "file.bin"),
        ("./file.bin", "file.bin"),
        ("path\\file", "path/file"),
        ("\\\\a\\b\\file.bin", "a/b/file.bin"),
        (".file", ".file"),
        ("a/../file.bin", "file.bin"),
        ("./a/./b/../c", "a/c"),
    ],
)
def test_testcase_21(path, expected_result):
    """test TestCase.sanitize_path()"""
    assert TestCase.sanitize_path(path) == expected_result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from pytest import mark

from .iomanager import IOManager


def test_iomanager_01():
    """test a simple IOManager"""
    with IOManager() as iom:
        assert iom.server_map is not None
        assert not iom.tests
        assert iom._generated == 0
        assert iom._report_size == 1
        assert iom._test is None


@mark.parametrize(
    "report_size, iters",
    [
        (1, 1),
        (1, 2),
        (2, 2),
        (2, 3),
    ],
)
def test_iomanager_02(report_size, iters):
    """test IOManager create_testcase(), commit() and purge()"""
    with IOManager(report_size=report_size) as iom:
        assert not iom.tests
        for current in range(1, iters + 1):
            tcase = iom.create_testcase("test-adapter", 10)
            assert iom._generated == current
            assert iom._test
            precommit_size = len(iom.tests)
            iom.commit()
            assert iom._test is None
            assert tcase == iom.tests[0]
            size = len(iom.tests)
            assert precommit_size <= size
            assert size <= report_size
        assert len(iom.tests) == report_size
        iom.purge()
        assert iom._test is None
        assert not iom.tests


def test_iomanager_03():
    """test IOManager.page_name()"""
    with IOManager() as iom:
        assert iom.page_name() != iom.page_name(offset=1)
        next_page = iom.page_name(offset=1)
        iom._generated += 1
        assert iom.page_name() == next_page


def test_iomanager_04():
    """test IOManager.create_testcase()"""
    time_limit = 10
    with IOManager() as iom:
        assert iom._generated == 0
        assert iom._report_size == 1
        assert not iom.tests
        assert not iom.server_map.dynamic
        assert not iom.server_map.include
        assert not iom.server_map.redirect
        tcase = iom.create_testcase("test-adapter", time_limit)
        assert tcase is not None
        assert not any(tcase.optional)
        assert tcase.time_limit == time_limit
        assert "grz_current_test" in iom.server_map.redirect
        assert iom.server_map.redirect["grz_current_test"].target == tcase.landing_page
        assert "grz_next_test" in iom.server_map.redirect
        assert iom._test is not None
        iom.purge()
        assert iom._test is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer beautify strategy definitions.

Constants:
    HAVE_CSSBEAUTIFIER (bool): True if `cssbeautifier` module is available.
    HAVE_JSBEAUTIFIER (bool): True if `jsbeautifier` module is available.
"""
import re
from abc import ABC, abstractmethod
from logging import getLogger

from lithium.testcases import TestcaseLine

try:
    import cssbeautifier
    import cssbeautifier.css.options

    HAVE_CSSBEAUTIFIER = True
except ImportError:  # pragma: no cover
    HAVE_CSSBEAUTIFIER = False
try:
    import jsbeautifier

    HAVE_JSBEAUTIFIER = True
except ImportError:  # pragma: no cover
    HAVE_JSBEAUTIFIER = False

from ...common.storage import TestCase
from . import Strategy, _contains_dd

LOG = getLogger(__name__)


def _split_normal_lines(data):
    """Like str.splitlines but only respect \n, \r\n, and \r .. leave other
    potential line break characters intact.

    Arguments:
        data (bytes): Input line to process.

    Returns:
        generator[bytes]: Yield lines split from data, not including line endings.
    """
    for win_line in data.split(b"\r\n"):
        for mac_line in win_line.split(b"\r"):
            yield from mac_line.split(b"\n")


class _BeautifyStrategy(Strategy, ABC):
    """A strategy that beautifies code in the testcase to make it more reducible.

    Implementors must define these class attributes:

    Class attributes:
        all_extensions (set(str)): Set of all file extensions to beautify.
        import_available (bool): Whether or not the beautify module was imported.
        import_name (str): The name of the beautify module imported (for error
                           reporting).
        name (str): The strategy name.
        native_extension (str): The native file extension for this type.
        tag_name (str): Tag name to search for in other (non-native) extensions.
    """

    all_extensions = None
    ignore_files = {"test_info.json", "prefs.js"}
    import_available = None
    import_name = None
    native_extension = None
    tag_name = None

    def __init__(self, testcases):
        """Initialize beautification strategy instance.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase)):
                List of testcases to reduce. The object does not take ownership of the
                testcases.
        """
        super().__init__(testcases)
        self._files_to_beautify = []
        for path in self._testcase_root.glob("**/*"):
            if (
                path.is_file()
                and path.suffix in self.all_extensions
                and path.name not in self.ignore_files
            ):
                if _contains_dd(path):
                    self._files_to_beautify.append(path)
        self._current_feedback = None
        tag_bytes = self.tag_name.encode("ascii")
        self._re_tag_start = re.compile(
            rb"<\s*" + tag_bytes + rb".*?>", flags=re.DOTALL | re.IGNORECASE
        )
        self._re_tag_end = re.compile(
            rb"</\s*" + tag_bytes + rb"\s*>", flags=re.IGNORECASE
        )

    @classmethod
    def sanity_check_cls_attrs(cls):
        """Sanity check the strategy class implementation.

        Raises:
            AssertionError: Required class attributes are missing or wrong type.

        Returns:
            None
        """
        super().sanity_check_cls_attrs()
        assert isinstance(cls.all_extensions, set)
        assert all(isinstance(ext, str) for ext in cls.all_extensions)
        assert isinstance(cls.import_available, bool)
        assert isinstance(cls.import_name, str)
        assert isinstance(cls.native_extension, str)
        assert isinstance(cls.tag_name, str)

    def update(self, success, served=None):
        """Inform the strategy whether or not the last beautification yielded was good.

        Arguments:
            success (bool): Whether or not the last beautification was acceptable.
            served (list(list(str))): The list of served files for each testcase in the
                                      last beautification.

        Returns:
            None
        """
        # beautify does nothing with served. it's unlikely a beautify operation alone
        # would render a file unserved.
        assert self._current_feedback is None
        self._current_feedback = success

    @classmethod
    @abstractmethod
    def beautify_bytes(cls, data):
        """Perform beautification on a code buffer.

        Arguments:
            data (bytes): The code data to be beautified.

        Returns:
            bytes: The beautified result.
        """

    def _chunks_to_beautify(self, before, to_beautify, file):
        """Iterate over `to_beautify` and find chunks of style/script to beautify.

        Arguments:
            before (bytes): The data preceding `to_beautify`. Used to check whether
                            `to_beautify` is already in an open <script> or <style> tag.
            to_beautify (bytes): The data to beautify.
            file (Path): The input file (used only to check if this is a .css/.js file)

        Yields:
            tuple (int,int): Slices of `to_beautify` that should be beautified.
        """
        # native extension, there's no need to search for tags
        if file.suffix == self.native_extension:
            yield (0, len(to_beautify))
            return

        # find the last <tag> preceding DDBEGIN
        last_tag_start = None
        for match in self._re_tag_start.finditer(before):
            last_tag_start = match
        in_tag_already = (
            # there was an open <tag>
            last_tag_start is not None
            # and it isn't followed by a closing </tag>
            and self._re_tag_end.search(before[last_tag_start.end(0) :]) is None
        )
        if in_tag_already:
            tag_end = self._re_tag_end.search(to_beautify)
            if tag_end is None:
                # similar to native case. DDBEGIN/END occurred inside the tag, so
                # no need to look further
                yield (0, len(to_beautify))
                return
            yield (0, tag_end.start(0))
            search_start = tag_end.end(0)
        else:
            search_start = 0

        # scan for <tag></tag> (with </tag> being optional for the last match)
        while True:
            tag_start = self._re_tag_start.search(to_beautify[search_start:])
            if tag_start is None:
                break
            chunk_start = search_start + tag_start.end(0)
            tag_end = self._re_tag_end.search(to_beautify[chunk_start:])
            if tag_end is None:
                # last </tag> was missing, stop looking
                yield (chunk_start, len(to_beautify))
                break
            yield (chunk_start, chunk_start + tag_end.start(0))
            search_start = chunk_start + tag_end.end(0)

    def __iter__(self):
        """Iterate over potential beautifications of testcases according to this
        strategy.

        The caller should evaluate each testcase set yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            list(grizzly.common.storage.TestCase): list of testcases with beautification
                                                   applied
        """
        if not self.import_available:
            LOG.warning("%s not available, skipping strategy.", self.import_name)
            return

        LOG.info("Beautifying %d files", len(self._files_to_beautify))
        for file_no, file in enumerate(self._files_to_beautify):
            LOG.info(
                "Beautifying %s (file %d/%d)",
                file.relative_to(self._testcase_root),
                file_no + 1,
                len(self._files_to_beautify),
            )

            # Use Lithium just to split the file at DDBEGIN/END.
            # Lithium already has the right logic for DDBEGIN/END.
            lith_tc = TestcaseLine()
            lith_tc.load(file)
            lith_tc.before = b"\n".join(_split_normal_lines(lith_tc.before))
            lith_tc.after = b"\n".join(_split_normal_lines(lith_tc.after))
            raw = b"\n".join(_split_normal_lines(b"".join(lith_tc.parts)))

            with file.open("wb") as testcase_fp:
                last = 0
                any_beautified = False
                testcase_fp.write(lith_tc.before)
                for start, end in self._chunks_to_beautify(lith_tc.before, raw, file):
                    before = raw[last:start]
                    testcase_fp.write(before)
                    to_beautify = raw[start:end]
                    LOG.debug("before: %r", to_beautify)
                    beautified = self.beautify_bytes(to_beautify)
                    LOG.debug("after: %r", beautified)
                    if beautified:
                        if before and not before.endswith(b"\n"):
                            beautified = b"\n" + beautified
                        if not beautified.endswith(b"\n"):
                            beautified = beautified + b"\n"

                        testcase_fp.write(beautified)
                        if beautified == to_beautify:
                            LOG.warning("Beautify had no effect!")
                        else:
                            any_beautified = True
                    elif to_beautify.strip():  # pragma: no cover
                        # this should never happen, but just in case...
                        # pragma: no cover
                        LOG.warning(
                            "No output from beautify! Writing %s unmodified.",
                            self.tag_name,
                        )
                        testcase_fp.write(to_beautify)  # pragma: no cover
                    last = end
                testcase_fp.write(raw[last:])
                testcase_fp.write(lith_tc.after)

                if last == 0:
                    LOG.warning("<%s> tags not found, skipping", self.tag_name)
                    continue

                if not any_beautified:
                    LOG.warning("Beautify had no effect on the file, skipping")
                    continue

            tc_hash = self._calculate_testcase_hash()
            if tc_hash in self._tried:
                LOG.info("cache hit, reverting")
                lith_tc.dump(file)
                continue

            yield TestCase.load(self._testcase_root, True)

            assert self._current_feedback is not None, "No feedback for last iteration"
            if self._current_feedback:
                LOG.info("%s was successful", self.name)
            else:
                LOG.warning("%s failed (reverting)", self.name)
                lith_tc.dump(file)
                self._tried.add(tc_hash)
            self._current_feedback = None


class CSSBeautify(_BeautifyStrategy):
    """Run CSS beautification on all CSS files and `<style>` tags.

    This should make the CSS more reducible if there are long lines with compound
    definitions.
    """

    all_extensions = {".css", ".htm", ".html", ".xhtml"}
    import_available = HAVE_CSSBEAUTIFIER
    import_name = "cssbeautifier"
    name = "cssbeautify"
    native_extension = ".css"
    opts = (
        ("end_with_newline", False),
        ("indent_size", 2),
        ("newline_between_rules", False),
        ("preserve_newlines", False),
    )
    tag_name = "style"

    @classmethod
    def beautify_bytes(cls, data):
        """Perform CSS beautification on a code buffer.

        Arguments:
            data (bytes): The code data to be beautified.

        Returns:
            bytes: The beautified result.
        """
        assert cls.import_available
        data = data.decode("utf-8", errors="surrogateescape")
        opts = cssbeautifier.css.options.BeautifierOptions(cls.opts)
        opts.eol = "\n"
        beautified = cssbeautifier.beautify(data, opts)
        return beautified.encode("utf-8", errors="surrogateescape")


class JSBeautify(_BeautifyStrategy):
    """Run JS beautification on all JS files and `<script>` tags.

    This should make the javascript more reducible if there are long lines with
    compound statements.
    """

    all_extensions = {".js", ".htm", ".html", ".xhtml"}
    import_available = HAVE_JSBEAUTIFIER
    import_name = "jsbeautifier"
    name = "jsbeautify"
    native_extension = ".js"
    opts = (
        ("end_with_newline", False),
        ("preserve_newlines", False),
    )
    tag_name = "script"
    try_catch_re = re.compile(r"(\s*try {)\r?\n\s*(.*)\r?\n\s*(}\s*catch.*)")

    @classmethod
    def beautify_bytes(cls, data):
        """Perform JS beautification on a code buffer.

        Arguments:
            data (bytes): The code data to be beautified.

        Returns:
            bytes: The beautified result.
        """
        assert HAVE_JSBEAUTIFIER
        data = data.decode("utf-8", errors="surrogateescape")
        opts = jsbeautifier.BeautifierOptions(cls.opts)
        opts.eol = "\n"
        beautified = jsbeautifier.beautify(data, opts)
        # All try/catch pairs will be expanded on their own lines
        # Collapse these pairs when only a single instruction is contained within
        beautified = cls.try_catch_re.sub(r"\1 \2 \3", beautified)
        return beautified.encode("utf-8", errors="surrogateescape")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
from collections import namedtuple
from itertools import chain, product
from os.path import normpath, split
from pathlib import Path
from shutil import copyfile, move, rmtree
from tempfile import NamedTemporaryFile, mkdtemp
from time import time
from zipfile import BadZipfile, ZipFile
from zlib import error as zlib_error

from ..target import AssetError, AssetManager
from .utils import grz_tmp

__all__ = ("TestCase", "TestCaseLoadFailure", "TestFileExists")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class TestCaseLoadFailure(Exception):
    """Raised when loading a TestCase fails"""


class TestFileExists(Exception):
    """Raised when adding a TestFile to a TestCase that has an existing
    TestFile with the same name"""


TestFile = namedtuple("TestFile", "file_name data_file")
TestFileMap = namedtuple("TestFileMap", "optional required")


class TestCase:
    __slots__ = (
        "adapter_name",
        "assets",
        "duration",
        "env_vars",
        "hang",
        "input_fname",
        "landing_page",
        "redirect_page",
        "time_limit",
        "timestamp",
        "_data_path",
        "_files",
    )

    def __init__(
        self,
        landing_page,
        redirect_page,
        adapter_name,
        input_fname=None,
        time_limit=None,
        timestamp=None,
    ):
        self.adapter_name = adapter_name
        self.assets = None
        self.duration = None
        self.env_vars = {}
        self.hang = False
        self.input_fname = input_fname  # file that was used to create the test case
        self.landing_page = self.sanitize_path(landing_page)
        if redirect_page is not None:
            self.redirect_page = self.sanitize_path(redirect_page)
        else:
            self.redirect_page = None
        self.time_limit = time_limit
        self.timestamp = time() if timestamp is None else timestamp
        self._files = TestFileMap(optional=[], required=[])
        self._data_path = Path(mkdtemp(prefix="testcase_", dir=grz_tmp("storage")))

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def add_batch(self, path, include_files, prefix=None, copy=True):
        """Iterate over files in include_files and attach the files that are
        located in path to TestCase.

        Args:
            path (str): Path to the root of the directory that contains files.
            include_files (iterable(str)): Paths of the files to be added to the
                                           TestCase if they exist in path.
            prefix (str): Path prefix to prepend to file when adding to
                          the TestCase.
            copy (bool): File will be copied if True otherwise the file will be moved.

        Returns:
            None
        """
        path = Path(path)
        for fname in include_files:
            file = Path(fname)
            try:
                relative = file.relative_to(path)
            except ValueError:
                # cannot add files outside path
                continue
            if prefix:
                relative = prefix / relative
            self.add_from_file(file, file_name=relative.as_posix(), copy=copy)

    def add_from_bytes(self, data, file_name, required=False):
        """Create a file and add it to the TestCase.

        Args:
            data (bytes): Data to write to file.
            file_name (str): Used as file path on disk and URI. Relative to wwwroot.
            required (bool): Indicates whether the file must be served.

        Returns:
            None
        """
        assert isinstance(data, bytes)
        with NamedTemporaryFile(delete=False, dir=grz_tmp("storage")) as in_fp:
            in_fp.write(data)
            data_file = Path(in_fp.name)

        try:
            self.add_from_file(
                data_file, file_name=file_name, required=required, copy=False
            )
        finally:
            # the temporary file should have been moved to the data path of the TestCase
            # unless an exception occurred so remove it if needed
            if data_file.is_file():
                data_file.unlink()

    def add_from_file(self, src_file, file_name=None, required=False, copy=False):
        """Add a file to the TestCase by either copying or moving an existing file.

        Args:
            src_file (str): Path to existing file to use.
            file_name (str): Used as file path on disk and URI. Relative to wwwroot.
                             If file_name is not given the name of the src_file
                             will be used.
            required (bool): Indicates whether the file must be served.
            copy (bool): File will be copied if True otherwise the file will be moved.

        Returns:
            None
        """
        src_file = Path(src_file)
        if file_name is None:
            file_name = src_file.name
        file_name = self.sanitize_path(file_name)

        test_file = TestFile(file_name, self._data_path / file_name)
        if test_file.file_name in self.contents:
            raise TestFileExists(f"{test_file.file_name!r} exists in test")

        test_file.data_file.parent.mkdir(parents=True, exist_ok=True)
        if copy:
            copyfile(src_file, test_file.data_file)
        else:
            move(src_file, test_file.data_file)

        # landing_page is always 'required'
        if required or test_file.file_name == self.landing_page:
            self._files.required.append(test_file)
        else:
            self._files.optional.append(test_file)

    def cleanup(self):
        """Remove all the test files.

        Args:
            None

        Returns:
            None
        """
        rmtree(self._data_path, ignore_errors=True)

    def clone(self):
        """Make a copy of the TestCase.

        Args:
            None

        Returns:
            TestCase: A copy of the TestCase instance.
        """
        result = type(self)(
            self.landing_page,
            self.redirect_page,
            self.adapter_name,
            self.input_fname,
            self.time_limit,
            self.timestamp,
        )
        result.assets = self.assets
        result.duration = self.duration
        result.env_vars = dict(self.env_vars)
        result.hang = self.hang

        # copy test data files
        for entry, required in chain(
            product(self._files.required, [True]),
            product(self._files.optional, [False]),
        ):
            result.add_from_file(
                entry.data_file, file_name=entry.file_name, required=required, copy=True
            )
        return result

    @property
    def contents(self):
        """All files in TestCase.

        Args:
            None

        Yields:
            str: File path (relative to wwwroot).
        """
        for tfile in chain(self._files.required, self._files.optional):
            yield tfile.file_name

    @property
    def data_path(self):
        """Location test data is stored on disk. This is intended to be used as wwwroot.

        Args:
            None

        Returns:
            Path: Directory containing test case files.
        """
        return self._data_path

    @property
    def data_size(self):
        """Total amount of data used (bytes) by the files in the TestCase.

        Args:
            None

        Returns:
            int: Total size of the TestCase in bytes.
        """
        total = 0
        for group in self._files:
            total += sum(x.data_file.stat().st_size for x in group)
        return total

    def dump(self, dst_path, include_details=False):
        """Write all the test case data to the filesystem.

        Args:
            dst_path (str): Path to directory to output data.
            include_details (bool): Output "test_info.json" file.

        Returns:
            None
        """
        dst_path = Path(dst_path)
        # save test files to dst_path
        for test_file in chain(self._files.required, self._files.optional):
            dst_file = dst_path / test_file.file_name
            dst_file.parent.mkdir(parents=True, exist_ok=True)
            copyfile(test_file.data_file, dst_file)
        # save test case files and meta data including:
        # adapter used, input file, environment info and files
        if include_details:
            assert isinstance(self.env_vars, dict)
            info = {
                "adapter": self.adapter_name,
                "duration": self.duration,
                "env": self.env_vars,
                "hang": self.hang,
                "input": Path(self.input_fname).name if self.input_fname else None,
                "target": self.landing_page,
                "time_limit": self.time_limit,
                "timestamp": self.timestamp,
            }
            # save target assets and update meta data
            if self.assets and not self.assets.is_empty():
                info["assets_path"] = "_assets_"
                info["assets"] = self.assets.dump(
                    str(dst_path), subdir=info["assets_path"]
                )
            with (dst_path / "test_info.json").open("w") as out_fp:
                json.dump(info, out_fp, indent=2, sort_keys=True)

    def get_file(self, path):
        """Lookup and return the TestFile with the specified file name.

        Args:
            path (str): Path (relative to wwwroot) of TestFile to retrieve.

        Returns:
            TestFile: TestFile with matching path otherwise None.
        """
        for tfile in chain(self._files.optional, self._files.required):
            if tfile.file_name == path:
                return tfile
        return None

    @classmethod
    def load(cls, path, adjacent=False):
        """Load TestCases from disk.

        Args:
            path (Path): Path can be:
                        1) A directory containing `test_info.json` and data.
                        2) A directory with one or more subdirectories of 1.
                        3) A zip archive containing testcase data or
                           subdirectories containing testcase data.
                        4) A single file to be used as a test case.
            adjacent (bool): Load adjacent files as part of the test case.
                             This is always the case when loading a directory.
                             WARNING: This should be used with caution!

        Returns:
            list: TestCases successfully loaded from path.
        """
        assert isinstance(path, Path)
        # unpack archive if needed
        if path.name.lower().endswith(".zip"):
            unpacked = mkdtemp(prefix="unpack_", dir=grz_tmp("storage"))
            try:
                with ZipFile(path) as zip_fp:
                    zip_fp.extractall(path=unpacked)
            except (BadZipfile, zlib_error):
                rmtree(unpacked, ignore_errors=True)
                raise TestCaseLoadFailure("Testcase archive is corrupted") from None
            path = Path(unpacked)
        else:
            unpacked = None
        # load testcase data from disk
        try:
            if path.is_file():
                tests = [cls.load_single(path, adjacent=adjacent)]
            elif path.is_dir():
                tests = []
                assets = None
                for tc_path in TestCase.scan_path(path):
                    tests.append(
                        cls.load_single(
                            tc_path, load_assets=assets is None, copy=unpacked is None
                        )
                    )
                    # only load assets once
                    if not assets and tests[-1].assets:
                        assets = tests[-1].assets
                # reuse AssetManager on all tests
                if assets:
                    for test in tests:
                        if test.assets is None:
                            test.assets = assets
                tests.sort(key=lambda tc: tc.timestamp)
            else:
                raise TestCaseLoadFailure("Invalid TestCase path")
        finally:
            if unpacked is not None:
                rmtree(unpacked, ignore_errors=True)
        return tests

    @classmethod
    def load_single(cls, path, adjacent=False, load_assets=True, copy=True):
        """Load contents of a TestCase from disk. If `path` is a directory it must
        contain a valid 'test_info.json' file.

        Args:
            path (Path): Path to the directory or file to load.
            adjacent (bool): Load adjacent files as part of the TestCase.
                             This is always true when loading a directory.
                             WARNING: This should be used with caution!
            load_assets (bool): Load assets files.
            copy (bool): Files will be copied if True otherwise the they will be moved.

        Returns:
            TestCase: A TestCase.
        """
        assert isinstance(path, Path)
        if path.is_dir():
            # load using test_info.json
            try:
                with (path / "test_info.json").open("r") as in_fp:
                    info = json.load(in_fp)
            except OSError:
                raise TestCaseLoadFailure("Missing 'test_info.json'") from None
            except ValueError:
                raise TestCaseLoadFailure("Invalid 'test_info.json'") from None
            if not isinstance(info.get("target"), str):
                raise TestCaseLoadFailure("'test_info.json' has invalid 'target' entry")
            entry_point = Path(path / info["target"])
            if not entry_point.is_file():
                raise TestCaseLoadFailure(
                    f"Entry point {info['target']!r} not found in {path}"
                )
            # always load all contents of a directory if a 'test_info.json' is loaded
            adjacent = True
        elif path.is_file():
            entry_point = path
            info = {}
        else:
            raise TestCaseLoadFailure(f"Missing or invalid TestCase '{path}'")
        # create testcase and add data
        test = cls(
            entry_point.relative_to(entry_point.parent).as_posix(),
            None,
            info.get("adapter", None),
            input_fname=info.get("input", None),
            time_limit=info.get("time_limit", None),
            timestamp=info.get("timestamp", 0),
        )
        test.duration = info.get("duration", None)
        test.hang = info.get("hang", False)
        test.add_from_file(
            entry_point, file_name=test.landing_page, required=True, copy=copy
        )
        if info:
            # load assets
            try:
                if load_assets and info.get("assets", None):
                    test.assets = AssetManager.load(
                        info.get("assets"),
                        str(entry_point.parent / info.get("assets_path", "")),
                    )
            except (AssetError, OSError) as exc:
                test.cleanup()
                raise TestCaseLoadFailure(str(exc)) from None
            # load environment variables
            test.env_vars = info.get("env", {})
            assert isinstance(test.env_vars, dict)
            # sanity check environment variable data
            for name, value in test.env_vars.items():
                if not isinstance(name, str) or not isinstance(value, str):
                    test.cleanup()
                    if test.assets:
                        test.assets.cleanup()
                    raise TestCaseLoadFailure("'env' contains invalid entries")
        # load all adjacent data from directory
        if adjacent:
            asset_path = info.get("assets_path", None)
            for entry in Path(entry_point.parent).rglob("*"):
                if not entry.is_file():
                    continue
                location = entry.relative_to(entry_point.parent).as_posix()
                # ignore asset path
                if asset_path and location.startswith(asset_path):
                    continue
                # ignore files that have been previously loaded
                if location in (test.landing_page, "test_info.json"):
                    continue
                # NOTE: when loading all files except the entry point are
                # marked as `required=False`
                test.add_from_file(
                    entry,
                    file_name=location,
                    required=False,
                    copy=copy,
                )
        return test

    @property
    def optional(self):
        """Get file paths of optional files.

        Args:
            None

        Yields:
            str: File path of each optional file.
        """
        for test in self._files.optional:
            yield test.file_name

    def pop_assets(self):
        """Remove AssetManager from TestCase.

        Args:
            None

        Returns:
            AssetManager: AssetManager if exists otherwise None.
        """
        if self.assets is None:
            assets = None
        else:
            assets = self.assets
            self.assets = None
        return assets

    def purge_optional(self, keep):
        """Remove optional files that are not in keep.

        Args:
            keep (iterable(str)): Files that will not be removed. This can contain
                                  absolute (includes) and relative paths.

        Returns:
            None
        """
        to_remove = []
        # iterate over optional files
        for idx, opt in enumerate(self._files.optional):
            # check entries in 'keep' for a match
            if not any(x.endswith(opt.file_name) for x in keep):
                to_remove.append(idx)
        # purge
        for idx in reversed(to_remove):
            self._files.optional.pop(idx).data_file.unlink()

    @staticmethod
    def sanitize_path(path):
        """Sanitize given path for use as a URI path.

        Args:
            path (str): Path to sanitize. Must be relative to wwwroot.

        Returns:
            str: Sanitized path.
        """
        assert isinstance(path, str)
        # check for missing filename or path containing drive letter (Windows)
        if split(path)[-1] in ("", ".", "..") or ":" in path:
            raise ValueError(f"invalid path {path!r}")
        # normalize path
        path = normpath(path).replace("\\", "/")
        # check normalized path does not resolve to location outside of '.'
        if path.startswith("../"):
            raise ValueError(f"invalid path {path!r}")
        return path.lstrip("/")

    @staticmethod
    def scan_path(path):
        """Check path and subdirectories for potential test cases. Subdirectories are
        only scanned one level deep.

        Args:
            path (Path): Path to scan.

        Yields:
            Path: Directory containing 'test_info.json' (potential testcase).
        """
        if (path / "test_info.json").is_file():
            yield path
        else:
            for entry in path.iterdir():
                if (entry / "test_info.json").is_file():
                    yield entry



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server worker
"""
from logging import getLogger
from re import compile as re_compile
from socket import SHUT_RDWR
from socket import error as sock_error
from socket import timeout as sock_timeout
from sys import exc_info
from threading import Thread, ThreadError, active_count
from time import sleep
from urllib.parse import quote, unquote, urlparse

from .server_map import Resource

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class WorkerError(Exception):
    """Raised by Worker"""


class Worker:
    DEFAULT_REQUEST_LIMIT = 0x1000  # 4KB
    DEFAULT_TX_SIZE = 0x10000  # 64KB
    REQ_PATTERN = re_compile(b"^GET\\s/(?P<url>\\S*)\\sHTTP/1")

    __slots__ = ("_conn", "_thread")

    def __init__(self, conn, thread):
        self._conn = conn
        self._thread = thread

    @staticmethod
    def _200_header(c_length, c_type, encoding="ascii"):
        assert c_type is not None
        data = (
            "HTTP/1.1 200 OK\r\n"
            "Cache-Control: max-age=0, no-cache\r\n"
            f"Content-Length: {c_length}\r\n"
            f"Content-Type: {c_type}\r\n"
            "Connection: close\r\n\r\n"
        )
        return data.encode(encoding)

    @staticmethod
    def _307_redirect(redirect_to, encoding="ascii"):
        data = (
            "HTTP/1.1 307 Temporary Redirect\r\n"
            f"Location: {redirect_to}\r\n"
            "Connection: close\r\n\r\n"
        )
        return data.encode(encoding)

    @staticmethod
    def _4xx_page(code, hdr_msg, close=-1, encoding="ascii"):
        if close < 0:
            content = f"<h3>{code}!</h3>"
        else:
            content = (
                "<script>\n"
                "window.onload = () => "
                f"{{ window.setTimeout(window.close, {close * 1000}) }}\n"
                "</script>\n"
                '<body style="background-color:#ffffe0">\n'
                f"<h3>{code}! - Calling window.close() in {close} seconds</h3>\n"
                "</body>\n"
            )
        data = (
            f"HTTP/1.1 {code} {hdr_msg}\r\n"
            f"Content-Length: {len(content)}\r\n"
            "Content-Type: text/html\r\n"
            f"Connection: close\r\n\r\n{content}"
        )
        return data.encode(encoding)

    def close(self):
        if not self.done:
            LOG.debug("closing socket while thread is running!")
            # shutdown socket to avoid hang
            self._conn.shutdown(SHUT_RDWR)
        self._conn.close()
        self.join(timeout=60)
        if self._thread is not None and self._thread.is_alive():
            # this is here to catch unexpected hangs
            raise WorkerError("Worker thread failed to join!")

    @property
    def done(self):
        if self._thread is not None and not self._thread.is_alive():
            self.join()
            self._thread = None
        return self._thread is None

    @classmethod
    def handle_request(cls, conn, serv_job):
        finish_job = False  # call finish() on return
        try:
            # receive all the incoming data
            raw_request = conn.recv(cls.DEFAULT_REQUEST_LIMIT)
            if not raw_request:
                LOG.debug("raw_request was empty")
                serv_job.accepting.set()
                return

            # check if request can be handled
            raw_url = cls.REQ_PATTERN.match(raw_request)
            if raw_url is None:
                serv_job.accepting.set()
                conn.sendall(cls._4xx_page(400, "Bad Request", serv_job.auto_close))
                LOG.debug(
                    "400 request length %d (%d to go)",
                    len(raw_request),
                    serv_job.pending,
                )
                return

            # lookup resource
            url = urlparse(unquote(raw_url.group("url").decode("ascii")))
            LOG.debug("lookup_resource(%r)", url.path)
            resource = serv_job.lookup_resource(url.path)
            if resource:
                if resource.type in (Resource.URL_FILE, Resource.URL_INCLUDE):
                    finish_job = serv_job.remove_pending(str(resource.target))
                elif resource.type in (Resource.URL_DYNAMIC, Resource.URL_REDIRECT):
                    finish_job = serv_job.remove_pending(url.path)
                else:  # pragma: no cover
                    # this should never happen
                    raise WorkerError(f"Unknown resource type {resource.type!r}")
                if serv_job.forever:
                    finish_job = False

            if not finish_job:
                serv_job.accepting.set()
            else:
                LOG.debug("expecting to finish")

            # send response
            if resource is None:
                conn.sendall(cls._4xx_page(404, "Not Found", serv_job.auto_close))
                LOG.debug("404 %r (%d to go)", url.path, serv_job.pending)
            elif resource.type == Resource.URL_REDIRECT:
                redirect_to = [quote(resource.target)]
                if url.query:
                    LOG.debug("appending query %r", url.query)
                    redirect_to.append(url.query)
                conn.sendall(cls._307_redirect("?".join(redirect_to)))
                LOG.debug(
                    "307 %r -> %r (%d to go)",
                    url.path,
                    resource.target,
                    serv_job.pending,
                )
            elif resource.type == Resource.URL_DYNAMIC:
                # pass query string to callback
                data = resource.target(url.query)
                if not isinstance(data, bytes):
                    LOG.debug("dynamic request: %r", url.path)
                    raise TypeError("dynamic request callback must return 'bytes'")
                conn.sendall(cls._200_header(len(data), resource.mime))
                conn.sendall(data)
                LOG.debug(
                    "200 %r - dynamic request (%d to go)", url.path, serv_job.pending
                )
            elif serv_job.is_forbidden(
                resource.target, is_include=resource.type == Resource.URL_INCLUDE
            ):
                # NOTE: this does info leak if files exist on disk.
                # We could replace 403 with 404 if it turns out we care.
                # However this is meant to only be accessible via localhost.
                LOG.debug("target %r", str(resource.target))
                conn.sendall(cls._4xx_page(403, "Forbidden", serv_job.auto_close))
                LOG.debug("403 %r (%d to go)", url.path, serv_job.pending)
            else:
                # serve the file
                LOG.debug("target %r", str(resource.target))
                data_size = resource.target.stat().st_size
                LOG.debug(
                    "sending: %s bytes, mime: %r", format(data_size, ","), resource.mime
                )
                with resource.target.open("rb") as in_fp:
                    conn.sendall(cls._200_header(data_size, resource.mime))
                    offset = 0
                    while offset < data_size:
                        conn.sendall(in_fp.read(cls.DEFAULT_TX_SIZE))
                        offset = in_fp.tell()
                LOG.debug("200 %r (%d to go)", url.path, serv_job.pending)
                serv_job.mark_served(resource.target)

        except (sock_error, sock_timeout):
            _, exc_obj, exc_tb = exc_info()
            LOG.debug("%r - line %r", exc_obj, exc_tb.tb_lineno if exc_tb else None)
            if not finish_job:
                serv_job.accepting.set()

        except Exception:  # pylint: disable=broad-except
            # set finish_job to abort immediately
            finish_job = True
            if serv_job.exceptions.empty():
                serv_job.exceptions.put(exc_info())

        finally:
            conn.close()
            if finish_job:
                serv_job.finish()
            serv_job.worker_complete.set()

    def join(self, timeout=None):
        if self._thread is not None:
            self._thread.join(timeout=timeout)
            if not self._thread.is_alive():
                self._thread = None

    @classmethod
    def launch(cls, listen_sock, job):
        assert job.accepting.is_set()
        conn = None
        try:
            conn, _ = listen_sock.accept()
            conn.settimeout(None)
            # create a worker thread to handle client request
            w_thread = Thread(target=cls.handle_request, args=(conn, job))
            job.accepting.clear()
            w_thread.start()
            return cls(conn, w_thread)
        except (sock_error, sock_timeout):
            if conn is not None:  # pragma: no cover
                conn.close()
        except ThreadError:
            if conn is not None:  # pragma: no cover
                conn.close()
            # reset accepting status
            job.accepting.set()
            LOG.warning("ThreadError (worker), threads: %d", active_count())
            # wait for system resources to free up
            sleep(0.1)
        return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from os.path import isfile
from pathlib import Path
from shutil import rmtree

from pytest import raises

from .assets import AssetError, AssetManager


def test_asset_manager_01(tmp_path):
    """test AssetManager()"""
    with AssetManager(base_path=str(tmp_path)) as assets:
        assert not assets.assets
        assert assets.path
        assert assets.is_empty()
        # add file (move)
        example = tmp_path / "example.txt"
        example.write_text("example")
        asset_location = assets.add("example_file", str(example), copy=False)
        assert str(Path(assets.path) / "example.txt") == asset_location
        assert len(assets.assets) == 1
        assert not assets.is_empty()
        assert not example.is_file()
        assert any(Path(assets.path).glob("**/example.txt"))
        # add existing asset - update asset (copy)
        example.write_text("example")
        assets.add("example_file", str(example))
        assert example.is_file()
        # add directory
        example = tmp_path / "example_path"
        example.mkdir()
        (example / "a").mkdir()
        (example / "a" / "1.txt").write_text("1")
        (example / "b").mkdir()
        (example / "b" / "2.txt").write_text("2")
        assets.add("example_path", str(example))
        rmtree(str(example))
        assert any(Path(assets.path).glob("**/example_path/a/1.txt"))
        assert any(Path(assets.path).glob("**/example_path/b/2.txt"))
        assert len(assets.assets) == 2
        # get
        assert "example.txt" in assets.get("example_file")
        assert "example_path" in assets.get("example_path")
        # remove directory
        assets.remove("example_path")
        assert len(assets.assets) == 1
        assert not any(Path(assets.path).glob("**/example_path"))
        # remove file
        assets.remove("example_file")
        assert len(assets.assets) == 0
        assert not any(Path(assets.path).iterdir())
        # cleanup
        assets.cleanup()
        assert not assets.assets
        assert assets.path is None


def test_asset_manager_02(tmp_path):
    """test AssetManager() failures"""
    with AssetManager(base_path=str(tmp_path)) as assets:
        # get missing
        assert assets.get("missing") is None
        # add missing file
        with raises(OSError, match="'missing' does not exist"):
            assets.add("a", "missing")
        assert not assets.assets
        # remove invalid asset
        assets.remove("missing")
        # add file
        example = tmp_path / "example.txt"
        example.write_text("example")
        assets.add("example_file", str(example), copy=True)
        # add existing file under as different asset (collision)
        with raises(AssetError, match="'example.txt' is an existing asset"):
            assets.add("collide", str(example))
        assert "collide" not in assets.assets
        example.unlink()
        # add file from existing asset
        example = Path(assets.path) / "example.txt"
        with raises(AssetError, match="Cannot add existing asset content"):
            assets.add("existing", str(example))
        assert "existing" not in assets.assets
        # add file from existing asset with asset name collision
        example = Path(assets.path) / "example.txt"
        with raises(AssetError, match="Cannot add existing asset content"):
            assets.add("example_file", str(example))
        assert "example_file" in assets.assets


def test_asset_manager_03(tmp_path):
    """test AssetManager() dump/load"""
    # test dump()
    with AssetManager(base_path=str(tmp_path)) as assets:
        # add file
        example = tmp_path / "example.txt"
        example.write_text("example")
        assets.add("example_file", str(example), copy=False)
        # add directory
        example = tmp_path / "example_path"
        example.mkdir()
        (example / "a").mkdir()
        (example / "a" / "1.txt").write_text("1")
        assets.add("example_path", str(example))
        # invalid entry
        assets.assets["invalid"] = "bad/path"
        # dump
        dump_path = tmp_path / "dump"
        dumped = assets.dump(str(dump_path), subdir="sub")
    assert len(dumped) == 2
    assert (dump_path / "sub" / "example.txt").is_file()
    assert (dump_path / "sub" / "example_path" / "a" / "1.txt").is_file()
    # test load()
    with AssetManager.load(
        dumped, str(dump_path / "sub"), base_path=str(tmp_path)
    ) as assets:
        assert len(assets.assets) == 2
        assert isfile(assets.assets["example_file"])


def test_asset_manager_04(tmp_path):
    """test AssetManager.add_batch()"""
    batch = []
    with AssetManager(base_path=str(tmp_path)) as assets:
        # add file
        example = tmp_path / "example.txt"
        example.write_text("example")
        batch.append(["example_file", str(example)])
        # add directory
        example = tmp_path / "example_path"
        example.mkdir()
        (example / "a").mkdir()
        (example / "a" / "1.txt").write_text("1")
        batch.append(["example_path", str(example)])
        assets.add_batch(batch)
        assert len(assets.assets) == 2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from itertools import chain
from logging import getLogger
from os import kill
from os.path import isfile
from pathlib import Path
from platform import system
from signal import SIGABRT

try:
    from signal import SIGUSR1
except ImportError:
    SIGUSR1 = None
from tempfile import TemporaryDirectory, mkdtemp
from time import sleep, time

from ffpuppet import BrowserTimeoutError, Debugger, FFPuppet, LaunchError, Reason
from ffpuppet.sanitizer_util import SanitizerOptions
from prefpicker import PrefPicker
from psutil import AccessDenied, NoSuchProcess, Process, process_iter, wait_procs

from ..common.reporter import Report
from ..common.utils import grz_tmp
from .target import Result, Target, TargetLaunchError, TargetLaunchTimeout
from .target_monitor import TargetMonitor

__all__ = ("PuppetTarget",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


class PuppetMonitor(TargetMonitor):
    def __init__(self, puppet):
        self._puppet = puppet

    def clone_log(self, log_id, offset=0):
        return self._puppet.clone_log(log_id, offset=offset)

    def is_running(self):
        return self._puppet.is_running()

    def is_healthy(self):
        return self._puppet.is_healthy()

    @property
    def launches(self):
        return self._puppet.launches

    def log_length(self, log_id):
        return self._puppet.log_length(log_id)


class PuppetTarget(Target):
    SUPPORTED_ASSETS = (
        # file containing line separated list of tokens to scan stderr/out for
        "abort-tokens",
        # xpi or directory containing the unpacked extension
        "extension",
        # LSan suppression list file
        "lsan-suppressions",
        # prefs.js file to use
        "prefs",
        # TSan suppression list file
        "tsan-suppressions",
        # UBSan suppression list file
        "ubsan-suppressions",
    )

    TRACKED_ENVVARS = (
        "ASAN_OPTIONS",
        "LSAN_OPTIONS",
        "TSAN_OPTIONS",
        "UBSAN_OPTIONS",
        "GNOME_ACCESSIBILITY",
        "MOZ_CHAOSMODE",
        "MOZ_FUZZ_CRASH_ON_LARGE_ALLOC",
        "MOZ_FUZZ_LARGE_ALLOC_LIMIT",
        "XPCOM_DEBUG_BREAK",
    )

    __slots__ = ("use_valgrind", "_extension", "_prefs", "_puppet")

    def __init__(self, binary, launch_timeout, log_limit, memory_limit, **kwds):
        super().__init__(
            binary,
            launch_timeout,
            log_limit,
            memory_limit,
            assets=kwds.pop("assets", None),
        )
        # TODO: clean up handling debuggers
        debugger = Debugger.NONE
        if kwds.pop("pernosco", False):
            debugger = Debugger.PERNOSCO
        if kwds.pop("rr", False):
            debugger = Debugger.RR
        if kwds.pop("valgrind", False):
            self.use_valgrind = True
            debugger = Debugger.VALGRIND
        self._extension = None
        self._prefs = None

        # create Puppet object
        self._puppet = FFPuppet(
            debugger=debugger,
            headless=kwds.pop("headless", None),
            working_path=str(grz_tmp("target")),
        )
        if kwds:
            LOG.warning(
                "PuppetTarget ignoring unsupported arguments: %s", ", ".join(kwds)
            )

    def _cleanup(self):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.clean_up()

    def close(self, force_close=False):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.close(force_close=force_close)

    @property
    def closed(self):
        return self._puppet.reason is not None

    def create_report(self, is_hang=False):
        logs = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
        self.save_logs(logs)
        return Report(Path(logs), self.binary, is_hang=is_hang)

    def filtered_environ(self):
        # remove context specific entries from environment
        filtered = dict(self.environ)
        opts = SanitizerOptions()
        # iterate over *SAN_OPTIONS entries
        for san in (x for x in filtered if x.endswith("SAN_OPTIONS")):
            opts.load_options(filtered[san])
            # remove entries specific to the current environment
            opts.pop("external_symbolizer_path")
            opts.pop("log_path")
            opts.pop("strip_path_prefix")
            opts.pop("suppressions")
            filtered[san] = str(opts)
        # remove empty entries
        return {k: v for k, v in filtered.items() if v}

    def is_idle(self, threshold):
        for _, cpu in self._puppet.cpu_usage():
            if cpu >= threshold:
                return False
        return True

    @property
    def monitor(self):
        if self._monitor is None:
            self._monitor = PuppetMonitor(self._puppet)
        return self._monitor

    def check_result(self, ignored):
        result = Result.NONE
        # check if there has been a crash, hangs will appear as SIGABRT
        if not self._puppet.is_healthy():
            self.close()
            # something has happened figure out what
            if self._puppet.reason == Reason.CLOSED:
                LOG.debug("target.close() was called")
            elif self._puppet.reason == Reason.EXITED:
                LOG.debug("target closed itself")
            elif (
                self._puppet.reason == Reason.WORKER
                and "memory" in ignored
                and "ffp_worker_memory_usage" in self._puppet.available_logs()
            ):
                result = Result.IGNORED
                LOG.debug("memory limit exceeded")
            elif (
                self._puppet.reason == Reason.WORKER
                and "log-limit" in ignored
                and "ffp_worker_log_size" in self._puppet.available_logs()
            ):
                result = Result.IGNORED
                LOG.debug("log size limit exceeded")
            else:
                # crash or hang (forced SIGABRT) has been detected
                LOG.debug("result detected (%s)", self._puppet.reason.name)
                result = Result.FOUND
        return result

    def handle_hang(self, ignore_idle=True):
        was_idle = False
        if self._puppet.is_healthy():
            proc_usage = sorted(self._puppet.cpu_usage(), key=lambda x: x[1])
            if proc_usage:
                pid, cpu = proc_usage.pop()
                if ignore_idle and cpu < 15:
                    # don't send SIGABRT if process is idle
                    LOG.debug("ignoring idle hang (%0.1f%%)", cpu)
                    was_idle = True
                elif system() == "Linux":
                    # sending SIGABRT is only supported on Linux for now
                    # TODO: add/test on other OSs
                    LOG.debug("sending SIGABRT to %r (%0.1f%%)", pid, cpu)
                    try:
                        kill(pid, SIGABRT)
                    except OSError:
                        LOG.warning("Failed to send SIGABRT to pid %d", pid)
                    self._puppet.wait(timeout=10)
        # always call close() since this function should only/always
        # be called when there has been a timeout
        self.close()
        return was_idle

    def dump_coverage(self, timeout=90):
        if system() != "Linux":
            LOG.debug("dump_coverage() only supported on Linux")
            return

        assert SIGUSR1 is not None
        pid = self._puppet.get_pid()
        if pid is None or not self._puppet.is_healthy():
            LOG.debug("Skipping coverage dump (target is not in a good state)")
            return
        # If at this point, the browser is in a good state, i.e. no crashes
        # or hangs, so signal the browser to dump coverage.
        running_procs = 0
        signaled_pids = []
        try:
            # send SIGUSR1 to browser processes
            parent_proc = Process(pid)
            for proc in chain([parent_proc], parent_proc.children(recursive=True)):
                # avoid sending SIGUSR1 to non-browser processes
                if Path(proc.exe()).name.startswith("firefox"):
                    LOG.debug(
                        "Sending SIGUSR1 to %d (%s)",
                        proc.pid,
                        "parent" if proc.pid == pid else "child",
                    )
                    try:
                        kill(proc.pid, SIGUSR1)
                        signaled_pids.append(proc.pid)
                    except OSError:
                        LOG.warning("Failed to send SIGUSR1 to pid %d", proc.pid)
                if proc.is_running():
                    running_procs += 1
        except (AccessDenied, NoSuchProcess):  # pragma: no cover
            pass
        if not signaled_pids:
            LOG.warning(
                "SIGUSR1 not sent, no browser processes found (%d process(es) running)",
                running_procs,
            )
            return
        start_time = time()
        gcda_found = False
        delay = 0.1
        # wait for processes to write .gcda files (typically takes <1 second)
        while True:
            for proc in process_iter(attrs=["pid", "open_files"]):
                # scan signaled processes for open .gcda files
                if (
                    proc.info["pid"] in signaled_pids
                    and proc.info["open_files"]
                    and any(x.path.endswith(".gcda") for x in proc.info["open_files"])
                ):
                    gcda_found = True
                    # collect pid of process with open .gcda file
                    gcda_open = proc.info["pid"]
                    break
            else:
                gcda_open = None
            elapsed = time() - start_time
            if gcda_found:
                if gcda_open is None:
                    # success
                    LOG.debug("gcda dump took %0.2fs", elapsed)
                    break
                if elapsed >= timeout:
                    # timeout waiting for .gnco file to be written
                    LOG.warning(
                        "gcda file open by pid %d after %0.2fs", gcda_open, elapsed
                    )
                    try:
                        kill(gcda_open, SIGABRT)
                        wait_procs([Process(gcda_open)], timeout=15)
                    except (AccessDenied, NoSuchProcess, OSError):  # pragma: no cover
                        pass
                    self.close()
                    break
                if delay < 1.0:
                    # increase delay to a maximum of 1 second
                    # it is increased when waiting for the .gcno files to be written
                    # this decreases the number of calls to process_iter()
                    delay = min(1.0, delay + 0.1)
            elif elapsed >= 10:
                # assume we missed the process writing .gcda files
                LOG.warning("No gcda files seen after %0.2fs", elapsed)
                break
            if not self._puppet.is_healthy():
                LOG.warning("Browser failure during dump_coverage()")
                break
            sleep(delay)

    def launch(self, location):
        # setup environment
        env_mod = dict(self.environ)
        # do not allow network connections to non local endpoints
        env_mod["MOZ_DISABLE_NONLOCAL_CONNECTIONS"] = "1"
        env_mod["MOZ_CRASHREPORTER_SHUTDOWN"] = "1"
        try:
            self._puppet.launch(
                self.binary,
                launch_timeout=self.launch_timeout,
                location=location,
                log_limit=self.log_limit,
                memory_limit=self.memory_limit,
                prefs_js=self._prefs,
                extension=self._extension,
                env_mod=env_mod,
            )
        except LaunchError as exc:
            self.close()
            if isinstance(exc, BrowserTimeoutError):
                raise TargetLaunchTimeout(str(exc)) from None
            raise TargetLaunchError(str(exc), self.create_report()) from None

    def log_size(self):
        return self._puppet.log_length("stderr") + self._puppet.log_length("stdout")

    def merge_environment(self, extra):
        output = dict(extra)
        if self.environ:
            # prioritize existing environment variables
            output.update(self.environ)
            # merge contents of *SAN_OPTIONS
            org = SanitizerOptions()
            out = SanitizerOptions()
            for san in ("ASAN", "LSAN", "TSAN", "UBSAN"):
                opts = "_".join((san, "OPTIONS"))
                org.load_options(self.environ.get(opts, ""))
                if not org:
                    # nothing to add from original
                    continue
                out.load_options(extra.get(opts, ""))
                for opt, value in org:  # pylint: disable=not-an-iterable
                    out.add(opt, value, overwrite=True)
                output[opts] = str(out)
        self.environ = output

    def process_assets(self):
        self._extension = self.assets.get("extension")
        self._prefs = self.assets.get("prefs")
        # generate temporary prefs.js with prefpicker
        if self._prefs is None:
            LOG.debug("using prefpicker to generate prefs.js")
            with TemporaryDirectory(dir=grz_tmp("target")) as tmp_path:
                prefs = Path(tmp_path) / "prefs.js"
                template = PrefPicker.lookup_template("browser-fuzzing.yml")
                PrefPicker.load_template(template).create_prefsjs(prefs)
                self._prefs = self.assets.add("prefs", str(prefs), copy=False)
        abort_tokens = self.assets.get("abort-tokens")
        if abort_tokens:
            LOG.debug("loading 'abort tokens' from %r", abort_tokens)
            with open(abort_tokens) as in_fp:
                for line in in_fp:
                    line = line.strip()
                    if line:
                        self._puppet.add_abort_token(line)

        # configure sanitizer suppressions
        opts = SanitizerOptions()
        for sanitizer in ("lsan", "tsan", "ubsan"):
            asset = f"{sanitizer}-suppressions"
            # load existing sanitizer options from environment
            var_name = f"{sanitizer.upper()}_OPTIONS"
            opts.load_options(self.environ.get(var_name, ""))
            if self.assets.get(asset):
                # use suppression file if provided as asset
                opts.add("suppressions", repr(self.assets.get(asset)), overwrite=True)
            elif opts.get("suppressions"):
                supp_file = opts.pop("suppressions")
                if SanitizerOptions.is_quoted(supp_file):
                    supp_file = supp_file[1:-1]
                if isfile(supp_file):
                    # use environment specified suppression file
                    LOG.debug("using %r from environment", asset)
                    opts.add(
                        "suppressions",
                        repr(self.assets.add(asset, supp_file)),
                        overwrite=True,
                    )
                else:
                    LOG.warning("Missing %s suppressions file %r", sanitizer, supp_file)
            else:
                LOG.debug("%r does not contain suppressions", var_name)
                continue
            # update sanitized *SAN_OPTIONS
            LOG.debug("updating suppressions in %r", var_name)
            self.environ[var_name] = str(opts)

    def save_logs(self, *args, **kwargs):
        self._puppet.save_logs(*args, **kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server
"""
from errno import EADDRINUSE
from logging import getLogger
from pathlib import Path
from random import randint
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOL_SOCKET, gethostname, socket
from time import sleep, time

from .connection_manager import ConnectionManager
from .job import Job, Served

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


LOG = getLogger(__name__)


class Sapphire:
    LISTEN_TIMEOUT = 0.25

    __slots__ = ("_auto_close", "_max_workers", "_socket", "_timeout")

    def __init__(
        self, allow_remote=False, auto_close=-1, max_workers=10, port=None, timeout=60
    ):
        self._auto_close = auto_close  # call 'window.close()' on 4xx error pages
        self._max_workers = max_workers  # limit worker threads
        self._socket = Sapphire._create_listening_socket(allow_remote, port)
        self._timeout = None
        self.timeout = timeout

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    @staticmethod
    def _create_listening_socket(remote, port=None, retries=20, timeout=LISTEN_TIMEOUT):
        """Create listening socket. Search for an open socket if needed and
        and configure the socket. If a specific port is unavailable or no
        available ports can be found socket.error will be raised.

        Args:
            remote (bool): Accept all (non-local) incoming connections.
            port (int): Port to listen on. If None is given a random port will
                        be used.
            retries (int): Number of additional attempts to configure the socket.
            timeout (float): Used to set socket timeout.

        Returns:
            socket: A listening socket.
        """
        assert retries >= 0
        assert timeout > 0
        addr = "0.0.0.0" if remote else "127.0.0.1"
        sock = None
        for retry in reversed(range(max(retries + 1, 1))):
            try:
                sock = socket(AF_INET, SOCK_STREAM)
                sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
                sock.settimeout(timeout)
                # find an unused port and avoid blocked ports
                # see: searchfox.org/mozilla-central/source/netwerk/base/nsIOService.cpp
                # highest blocked port is 10080 (0x2760)
                sock.bind((addr, port or randint(0x2770, 0xFFFF)))
                sock.listen(5)
            except OSError as soc_e:
                if sock is not None:
                    sock.close()
                    sock = None
                if retry > 1 and soc_e.errno in (EADDRINUSE, 10013):
                    sleep(0.1)
                    continue
                raise
            break
        assert sock is not None
        return sock

    def clear_backlog(self):
        """Remove all pending connections from backlog. This should only be
        called when there isn't anything actively trying to connect.

        Args:
            None

        Returns:
            None
        """
        LOG.debug("clearing socket backlog")
        self._socket.settimeout(0)
        deadline = time() + 10
        while True:
            try:
                self._socket.accept()[0].close()
            except BlockingIOError:
                break
            except OSError as exc:
                LOG.debug("Error closing socket: %r", exc)
            else:
                LOG.debug("pending socket closed")
            # if this fires something is likely actively trying to connect
            assert deadline > time()
        self._socket.settimeout(self.LISTEN_TIMEOUT)

    def close(self):
        """Close listening server socket.

        Args:
            None

        Returns:
            None
        """
        if self._socket is not None:
            self._socket.close()

    @property
    def port(self):
        """Port number of listening socket.

        Args:
            None

        Returns:
            int: Listening port number.
        """
        return self._socket.getsockname()[1]

    def serve_path(
        self,
        path,
        continue_cb=None,
        forever=False,
        optional_files=None,
        server_map=None,
    ):
        """Serve files in path. On completion a list served files and a status
        code will be returned.
        The status codes include:
            - Served.ALL: All files excluding files in optional_files were served
            - Served.NONE: No files were served
            - Served.REQUEST: Some files were requested

        Args:
            path (Path): Directory to use as wwwroot.
            continue_cb (callable): A callback that can be used to exit the serve loop.
                                    This must be a callable that returns a bool.
            forever (bool): Continue to handle requests even after all files have
                            been served. This is meant to be used with continue_cb.
            optional_files (list(str)): Files that do not need to be served in order
                                        to exit the serve loop.
            server_map (ServerMap):

        Returns:
            tuple(int, tuple(str)): Status code and files served.
        """
        assert isinstance(path, Path)
        LOG.debug("serving '%s' (forever=%r)", path, forever)
        job = Job(
            path,
            auto_close=self._auto_close,
            forever=forever,
            optional_files=optional_files,
            server_map=server_map,
        )
        if not job.pending:
            job.finish()
            LOG.debug("nothing to serve")
            return (Served.NONE, tuple())
        with ConnectionManager(job, self._socket, self._max_workers) as loadmgr:
            was_timeout = not loadmgr.wait(self.timeout, continue_cb=continue_cb)
        LOG.debug("%s, timeout: %r", job.status, was_timeout)
        return (Served.TIMEOUT if was_timeout else job.status, tuple(job.served))

    @property
    def timeout(self):
        """The amount of time that must pass before exiting the serve loop and
        indicating a timeout.

        Args:
            None

        Returns:
            int: Timeout in seconds.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """The amount of time that must pass before exiting the serve loop and
        indicating a timeout.

        Args:
            value (int): Timeout in seconds.

        Returns:
            None
        """
        assert value >= 0
        self._timeout = value

    @classmethod
    def main(cls, args):
        try:
            with cls(
                allow_remote=args.remote, port=args.port, timeout=args.timeout
            ) as serv:
                LOG.info(
                    "Serving '%s' @ http://%s:%d/",
                    args.path.resolve(),
                    gethostname() if args.remote else "127.0.0.1",
                    serv.port,
                )
                status = serv.serve_path(args.path)[0]
            if status == Served.ALL:
                LOG.info("All test case content was served")
            else:
                LOG.warning("Failed to serve all test content")
        except KeyboardInterrupt:
            LOG.warning("Ctrl+C detected. Shutting down...")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server job
"""
from collections import namedtuple
from enum import Enum, unique
from logging import getLogger
from mimetypes import guess_type
from os.path import splitext
from pathlib import Path
from queue import Queue
from threading import Event, Lock

from .server_map import Resource

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


@unique
class Served(Enum):
    """Server Job status codes"""

    # all expected requests for required files have been received
    ALL = 0
    # no requests for required files have been received
    NONE = 1
    # some requests for required files have been received
    REQUEST = 2
    # timeout occurred
    TIMEOUT = 3


Tracker = namedtuple("Tracker", "files lock")


class Job:
    # MIME_MAP is used to support new or uncommon mime types.
    # Definitions in here take priority over mimetypes.guess_type().
    MIME_MAP = {
        ".avif": "image/avif",
        ".bmp": "image/bmp",
        ".ico": "image/x-icon",
        ".wave": "audio/x-wav",
        ".webp": "image/webp",
    }

    __slots__ = (
        "_complete",
        "_pending",
        "_served",
        "_wwwroot",
        "auto_close",
        "accepting",
        "exceptions",
        "forever",
        "server_map",
        "worker_complete",
    )

    def __init__(
        self,
        wwwroot,
        auto_close=-1,
        forever=False,
        optional_files=None,
        server_map=None,
    ):
        self._complete = Event()
        self._pending = Tracker(files=set(), lock=Lock())
        self._served = Tracker(files=set(), lock=Lock())
        self._wwwroot = wwwroot.resolve()
        self.accepting = Event()
        self.accepting.set()
        self.auto_close = auto_close
        self.exceptions = Queue()
        self.forever = forever
        self.server_map = server_map
        self.worker_complete = Event()
        self._build_queue(optional_files)

    def _build_queue(self, optional_files):
        # build file list to track files that must be served
        # this is intended to only be called once by __init__()
        for entry in self._wwwroot.rglob("*"):
            if not entry.is_file():
                continue
            location = entry.relative_to(self._wwwroot).as_posix()
            # do not add optional files to queue of required files
            if optional_files and location in optional_files:
                LOG.debug("optional: %r", location)
                continue
            if "?" in str(entry):
                LOG.warning(
                    "Cannot add files with '?' in path. Skipping %r", str(entry)
                )
                continue
            self._pending.files.add(str(entry.resolve()))
            LOG.debug("required: %r", location)
        # if nothing was found check if the path exists
        if not self._pending.files and not self._wwwroot.is_dir():
            raise OSError(f"'{self._wwwroot}' does not exist")
        if self.server_map:
            for redirect, resource in self.server_map.redirect.items():
                if resource.required:
                    self._pending.files.add(redirect)
                LOG.debug(
                    "%s: %r -> %r",
                    "required" if resource.required else "optional",
                    redirect,
                    resource.target,
                )
            for dyn_resp, resource in self.server_map.dynamic.items():
                if resource.required:
                    self._pending.files.add(dyn_resp)
                    LOG.debug("%s: %r -> %r", "required", dyn_resp, resource.target)
        LOG.debug("%d files required to serve", len(self._pending.files))

    @classmethod
    def lookup_mime(cls, url):
        mime = cls.MIME_MAP.get(splitext(url)[-1].lower())
        if mime is None:
            # default to "application/octet-stream"
            mime = guess_type(url)[0] or "application/octet-stream"
        return mime

    def lookup_resource(self, path):
        # check if path is a file in wwwroot
        try:
            local = self._wwwroot / path
            if local.is_file():
                local = local.resolve()
                with self._pending.lock:
                    required = str(local) in self._pending.files
                return Resource(
                    Resource.URL_FILE,
                    local,
                    mime=self.lookup_mime(path),
                    required=required,
                )
        except ValueError:  # pragma: no cover
            # this is for compatibility with python versions < 3.8
            # is_file() will raise if the path contains characters unsupported
            # at the OS level
            pass
        # look for path in server map
        if self.server_map is not None:
            if path in self.server_map.redirect:
                return self.server_map.redirect[path]
            if path in self.server_map.dynamic:
                return self.server_map.dynamic[path]
            # search include paths for a match
            for inc in (x for x in self.server_map.include if path.startswith(x)):
                LOG.debug("checking include %r", inc)
                file = path[len(inc) :].lstrip("/")
                local = Path(self.server_map.include[inc].target) / file
                try:
                    if not local.is_file():
                        continue
                except ValueError:  # pragma: no cover
                    # python versions < 3.8 compatibility
                    continue
                # file exists, look up resource
                return Resource(
                    Resource.URL_INCLUDE,
                    local.resolve(),
                    mime=self.server_map.include[inc].mime or self.lookup_mime(file),
                    required=self.server_map.include[inc].required,
                )
        return None

    def finish(self):
        self._complete.set()

    def mark_served(self, path):
        # update list of served files
        assert isinstance(path, Path)
        with self._served.lock:
            self._served.files.add(path)

    def is_complete(self, wait=None):
        if wait is not None:
            return self._complete.wait(wait)
        return self._complete.is_set()

    def is_forbidden(self, target, is_include=False):
        # it is assumed that these file exists on disk
        # and that the paths are absolute and sanitized
        target = str(target)
        if not is_include:
            # check if target is in wwwroot
            if target.startswith(str(self._wwwroot)):
                return False
        elif self.server_map:
            # check if target is in an included path
            for resource in self.server_map.include.values():
                if target.startswith(resource.target):
                    # target is in a valid include path
                    return False
        return True

    @property
    def pending(self):
        # number of pending files
        with self._pending.lock:
            return len(self._pending.files)

    def remove_pending(self, file_name):
        # return True when all file have been removed
        with self._pending.lock:
            if self._pending.files:
                self._pending.files.discard(file_name)
            return not self._pending.files

    @property
    def served(self):
        # served files
        # files served from wwwroot will have a relative path
        # include files will have an absolute path
        with self._served.lock:
            # make a copy of what is available (maybe a copy not necessary?)
            served = tuple(self._served.files)
        for path in served:
            try:
                # file is in wwwroot
                yield path.relative_to(self._wwwroot).as_posix()
            except ValueError:
                # include file
                yield path.as_posix()

    @property
    def status(self):
        with self._pending.lock:
            if not self._served.files:
                return Served.NONE
            if not self._pending.files:
                return Served.ALL
        return Served.REQUEST



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

if __name__ == "__main__":
    from ..replay.bucket import bucket_main
    from .args import ReduceFuzzManagerIDQualityArgs
    from .crash import main

    raise SystemExit(bucket_main(ReduceFuzzManagerIDQualityArgs().parse_args(), main))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .assets import AssetError, AssetManager
from .target import Result, Target, TargetError, TargetLaunchError, TargetLaunchTimeout

__all__ = (
    "AssetError",
    "AssetManager",
    "Result",
    "Target",
    "TargetError",
    "TargetLaunchError",
    "TargetLaunchTimeout",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod
from enum import Enum, unique
from logging import getLogger
from os import environ
from os.path import isfile
from threading import Lock

from ..common.utils import grz_tmp
from .assets import AssetManager

__all__ = ("Result", "Target", "TargetError", "TargetLaunchError")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


@unique
class Result(Enum):
    """Target results codes"""

    NONE = 0
    FOUND = 1
    IGNORED = 2


class TargetError(Exception):
    """Raised by Target"""


class TargetLaunchError(TargetError):
    """Raised if a failure during launch occurs"""

    def __init__(self, message, report):
        super().__init__(message)
        self.report = report


class TargetLaunchTimeout(TargetError):
    """Raised if the target does not launch within the defined amount of time"""


class Target(metaclass=ABCMeta):
    SUPPORTED_ASSETS = None
    TRACKED_ENVVARS = ()

    __slots__ = (
        "_assets",
        "_lock",
        "_monitor",
        "binary",
        "environ",
        "launch_timeout",
        "log_limit",
        "memory_limit",
    )

    def __init__(self, binary, launch_timeout, log_limit, memory_limit, assets=None):
        assert launch_timeout > 0
        assert log_limit >= 0
        assert memory_limit >= 0
        assert binary is not None and isfile(binary)
        assert assets is None or isinstance(assets, AssetManager)
        self._assets = assets if assets else AssetManager(base_path=grz_tmp("target"))
        self._lock = Lock()
        self._monitor = None
        self.binary = binary
        self.environ = self.scan_environment(dict(environ), self.TRACKED_ENVVARS)
        self.launch_timeout = launch_timeout
        self.log_limit = log_limit
        self.memory_limit = memory_limit

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    @property
    def assets(self):
        return self._assets

    @assets.setter
    def assets(self, assets):
        self._assets.cleanup()
        assert isinstance(assets, AssetManager)
        self._assets = assets

    @abstractmethod
    def _cleanup(self):
        pass

    @abstractmethod
    def check_result(self, ignored):
        pass

    def cleanup(self):
        # call target specific _cleanup first
        self._cleanup()
        self._assets.cleanup()

    @abstractmethod
    def close(self, force_close=False):
        pass

    @property
    @abstractmethod
    def closed(self):
        pass

    @abstractmethod
    def create_report(self, is_hang=False):
        pass

    def dump_coverage(self, _timeout=0):
        LOG.warning("dump_coverage() is not supported!")

    def filtered_environ(self):
        """Used to collect the environment to add to a testcase"""
        return dict(self.environ)

    @abstractmethod
    def handle_hang(self, ignore_idle=True):
        pass

    # TODO: move to monitor?
    def is_idle(self, _threshold):
        LOG.debug("Target.is_idle() not implemented! returning False")
        return False

    @abstractmethod
    def launch(self, location):
        pass

    def log_size(self):
        LOG.debug("log_size() not implemented! returning 0")
        return 0

    @abstractmethod
    def merge_environment(self, extra):
        pass

    @property
    @abstractmethod
    def monitor(self):
        pass

    @abstractmethod
    def process_assets(self):
        pass

    def reverse(self, remote, local):
        # remote->device, local->desktop
        pass

    @staticmethod
    def scan_environment(to_scan, tracked):
        # scan environment for tracked environment variables
        env = {}
        if tracked:
            for var in tracked:
                if var in to_scan:
                    env[var] = to_scan[var]
        return env

    @abstractmethod
    def save_logs(self, *args, **kwargs):
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from enum import IntEnum, unique
from logging import DEBUG, basicConfig, getLogger
from math import ceil
from os import getenv
from pathlib import Path
from tempfile import gettempdir

__all__ = (
    "ConfigError",
    "configure_logging",
    "display_time_limits",
    "DEFAULT_TIME_LIMIT",
    "Exit",
    "grz_tmp",
    "time_limits",
    "TIMEOUT_DELAY",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

DEFAULT_TIME_LIMIT = 30
GRZ_TMP = Path(getenv("GRZ_TMP", gettempdir()), "grizzly")
LOG = getLogger(__name__)
# TIMEOUT_DELAY is added to the test time limit to create the default timeout
TIMEOUT_DELAY = 15


class ConfigError(Exception):
    """Raised to indicate invalid configuration a state"""

    def __init__(self, message, exit_code):
        super().__init__(message)
        self.exit_code = exit_code


@unique
class Exit(IntEnum):
    """Exit codes"""

    SUCCESS = 0
    # unexpected error occurred (invalid input, unhanded exception, etc)
    ERROR = 1
    # invalid argument
    ARGS = 2
    # run aborted (ctrl+c, etc)
    ABORT = 3
    # unrelated Target failure (browser startup crash, etc)
    LAUNCH_FAILURE = 4
    # expected results not reproduced (opposite of SUCCESS)
    FAILURE = 5


def configure_logging(log_level):
    """Configure log output level and formatting.

    Args:
        log_level (int): Set log level.

    Returns:
        None
    """
    # allow force enabling log_level via environment
    if getenv("DEBUG", "0").lower() in ("1", "true"):
        log_level = DEBUG
    if log_level == DEBUG:
        date_fmt = None
        log_fmt = "%(asctime)s %(levelname).1s %(name)s | %(message)s"
    else:
        date_fmt = "%Y-%m-%d %H:%M:%S"
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt=date_fmt, level=log_level)


def display_time_limits(time_limit, timeout, no_harness):
    """Output configuration of time limits and harness.

    Args:
        time_limit (int): Time in seconds before harness attempts to close current test.
        timeout (int): Time in seconds before iteration is considered a timeout.
        no_harness (bool): Indicate whether harness will is disabled.

    Returns:
        None
    """
    if timeout > 0:
        if no_harness:
            LOG.info("Using timeout: %ds, harness: DISABLED", timeout)
        else:
            LOG.info("Using time limit: %ds, timeout: %ds", time_limit, timeout)
            if time_limit == timeout:
                LOG.info("To avoid unnecessary relaunches set timeout > time limit")
    else:
        if no_harness:
            LOG.info("Using timeout: DISABLED, harness: DISABLED")
        else:
            LOG.info("Using time limit: %ds, timeout: DISABLED,", time_limit)
        LOG.warning("TIMEOUT DISABLED, not recommended for automation")


def grz_tmp(*subdir):
    path = Path(GRZ_TMP, *subdir)
    path.mkdir(parents=True, exist_ok=True)
    return path


def time_limits(
    time_limit,
    timeout,
    tests=None,
    default_limit=DEFAULT_TIME_LIMIT,
    timeout_delay=TIMEOUT_DELAY,
):
    """Determine the test time limit and timeout. If time_limit or timeout is None
    it is calculated otherwise the provided value is used.

    Args:
        time_limit (int): Test time limit.
        timeout (int): Iteration timeout.
        tests (iterable): Testcases that may contain time limit values.
        default_limit (int): Value to used as default time limit.
        timeout_delay (int): Value to used as delay when calculating timeout.

    Returns:
        tuple (int, int): Time limit and timeout.
    """
    assert default_limit > 0
    assert timeout_delay >= 0
    # calculate time limit
    calc_limit = time_limit is None
    if calc_limit:
        # use default_limit as a minimum
        test_limits = [default_limit]
        if tests:
            test_limits.extend(int(ceil(x.duration)) for x in tests if x.duration)
        time_limit = max(test_limits)
    assert time_limit > 0
    # calculate timeout
    calc_timeout = timeout is None
    if calc_timeout:
        timeout = time_limit + timeout_delay
    elif calc_limit and time_limit > timeout > 0:
        LOG.debug("calculated time limit > given timeout, using timeout")
        time_limit = timeout
    assert timeout >= 0
    # timeout should always be >= time limit unless timeout is disabled
    assert timeout >= time_limit or timeout == 0
    return time_limit, timeout



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
unit tests for grizzly.replay.main
"""
from pytest import mark

from sapphire import Served

from ..common.reporter import Report
from ..common.storage import TestCase, TestCaseLoadFailure
from ..common.utils import Exit
from ..target import (
    AssetManager,
    Result,
    Target,
    TargetLaunchError,
    TargetLaunchTimeout,
)
from .replay import ReplayManager
from .test_replay import _fake_save_logs

pytestmark = mark.usefixtures(
    "patch_collector",
    "tmp_path_grz_tmp",
)


def test_main_01(mocker, tmp_path):
    """test ReplayManager.main()"""
    # This is a typical scenario - a test that reproduces results ~50% of the time.
    # Of the four attempts only the first and third will 'reproduce' the result
    # and the forth attempt should be skipped.
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    # mock Sapphire.serve_path only
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path",
        autospec=True,
        return_value=(Served.ALL, ["test.html"]),  # passed to Target.check_result
    )
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    target = mocker.Mock(spec_set=Target, binary="bin", environ={}, launch_timeout=30)
    target.assets = mocker.Mock(spec_set=AssetManager)
    target.check_result.side_effect = (Result.FOUND, Result.NONE, Result.FOUND)
    target.filtered_environ.return_value = {"ENV": "123"}
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    with TestCase("test.html", None, "adpt") as src:
        src.env_vars["TEST_VAR"] = "100"
        src.add_from_bytes(b"test", "test.html")
        src.dump(str(tmp_path / "testcase"), include_details=True)
    # setup args
    log_path = tmp_path / "logs"
    (tmp_path / "sig.json").write_bytes(
        b'{"symptoms": [{"type": "crashAddress", "address": "0"}]}'
    )
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=tmp_path / "testcase",
        launch_attempts=3,
        logs=log_path,
        min_crashes=2,
        no_harness=False,
        pernosco=False,
        post_launch_delay=None,
        relaunch=1,
        repeat=4,
        rr=False,
        sig=tmp_path / "sig.json",
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.SUCCESS
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 3
    assert target.check_result.call_count == 3
    assert target.merge_environment.call_count == 1
    assert serve_path.call_count == 3
    assert load_target.call_count == 1
    assert target.close.call_count == 4
    assert target.filtered_environ.call_count == 2
    assert target.cleanup.call_count == 1
    assert target.assets.add.call_count == 0
    assert target.assets.is_empty.call_count == 1
    assert log_path.is_dir()
    assert any(log_path.glob("reports/*/log_asan_blah.txt"))
    assert any(log_path.glob("reports/*/log_stderr.txt"))
    assert any(log_path.glob("reports/*/log_stdout.txt"))


@mark.parametrize(
    "repro_results,",
    [
        # no results
        (Result.NONE, Result.NONE),
        # results do not match signature
        (Result.FOUND, Result.FOUND),
        # no result and failed signature match
        (Result.FOUND, Result.NONE),
    ],
)
def test_main_02(mocker, tmp_path, repro_results):
    """test ReplayManager.main() - no repro"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    # mock Sapphire.serve_path only
    mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path",
        autospec=True,
        return_value=(Served.ALL, ["test.html"]),  # passed to Target.check_result
    )
    # setup Target
    target = mocker.Mock(spec_set=Target, binary="bin", environ={}, launch_timeout=30)
    target.check_result.side_effect = repro_results
    target.filtered_environ.return_value = {}
    target.save_logs = _fake_save_logs
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec=Target, return_value=target),
    )
    (tmp_path / "sig.json").write_bytes(
        b'{"symptoms": [{"type": "stackFrames", "functionNames": ["no-match"]}]}'
    )
    # setup args
    (tmp_path / "test.html").touch()
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        input=tmp_path / "test.html",
        launch_attempts=3,
        logs=tmp_path / "logs",
        min_crashes=2,
        no_harness=True,
        pernosco=False,
        post_launch_delay=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=tmp_path / "sig.json",
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.FAILURE
    assert target.check_result.call_count == 1
    assert not target.environ
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1


def test_main_03(mocker):
    """test ReplayManager.main() error cases"""
    fake_sig = mocker.patch("grizzly.replay.replay.CrashSignature", autospec=True)
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    fake_load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    fake_tc = mocker.patch("grizzly.replay.replay.TestCase", autospec=True)
    # setup args
    args = mocker.Mock(
        asset=[],
        ignore=[],
        input="test",
        min_crashes=1,
        no_harness=True,
        pernosco=False,
        post_launch_delay=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    # user abort
    fake_load_target.side_effect = KeyboardInterrupt
    assert ReplayManager.main(args) == Exit.ABORT
    fake_load_target.reset_mock()
    # invalid test case
    fake_tc.load.side_effect = TestCaseLoadFailure
    assert ReplayManager.main(args) == Exit.ERROR
    assert fake_load_target.call_count == 0
    # no test cases
    fake_tc.load.side_effect = None
    fake_tc.load.return_value = []
    assert ReplayManager.main(args) == Exit.ERROR
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # multiple test cases with --no-harness
    fake_tc.load.return_value = [
        mocker.Mock(spec_set=TestCase, env_vars={}, hang=False) for _ in range(2)
    ]
    assert ReplayManager.main(args) == Exit.ARGS
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # signature required replaying hang
    fake_tc.load.return_value = [mocker.Mock(spec_set=TestCase, env_vars={}, hang=True)]
    assert ReplayManager.main(args) == Exit.ERROR
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # can't ignore timeout replaying hang
    args.ignore = ["timeout"]
    args.sig = "sig"
    fake_tc.load.return_value = [mocker.Mock(spec_set=TestCase, env_vars={}, hang=True)]
    assert ReplayManager.main(args) == Exit.ERROR
    assert fake_sig.fromFile.call_count == 1
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()


def test_main_04(mocker):
    """test ReplayManager.main() target exceptions"""
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    reporter = mocker.patch("grizzly.replay.replay.FailedLaunchReporter", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    mocker.patch("grizzly.replay.replay.TestCase", autospec=True)
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(
            spec_set=Target,
            return_value=mocker.NonCallableMock(spec_set=Target, launch_timeout=30),
        ),
    )
    # setup args
    args = mocker.MagicMock(
        input="test",
        min_crashes=1,
        no_harness=True,
        post_launch_delay=None,
        pernosco=False,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    # target launch error
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run",
        side_effect=TargetLaunchError("", mocker.Mock(spec_set=Report)),
    )
    assert ReplayManager.main(args) == Exit.LAUNCH_FAILURE
    assert reporter.return_value.submit.call_count == 1
    reporter.reset_mock()
    # target launch timeout
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run", side_effect=TargetLaunchTimeout
    )
    assert ReplayManager.main(args) == Exit.LAUNCH_FAILURE
    assert reporter.return_value.submit.call_count == 0


def test_main_05(mocker, tmp_path):
    """test ReplayManager.main() loading specified assets"""
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path",
        autospec=True,
        return_value=(None, ["test.html"]),  # passed to Target.check_result
    )
    # setup Target
    target = mocker.NonCallableMock(spec_set=Target, binary="bin", launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.filtered_environ.return_value = {}
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec_set=Target, return_value=target),
    )
    asset = tmp_path / "sample_asset"
    asset.touch()
    # setup args
    args = mocker.Mock(
        asset=[["from_cmdline", str(asset)]],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        launch_attempts=3,
        min_crashes=1,
        no_harness=True,
        pernosco=False,
        post_launch_delay=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=1,
        timeout=None,
        valgrind=False,
    )
    log_path = tmp_path / "logs"
    args.logs = log_path
    input_path = tmp_path / "input"
    input_path.mkdir()
    # build a test case
    entry_point = input_path / "test.html"
    entry_point.touch()
    with TestCase("test.html", None, "test-adapter") as src:
        src.add_from_file(str(entry_point))
        src.dump(str(input_path), include_details=True)
    args.input = input_path
    with AssetManager(base_path=str(tmp_path)) as assets:
        target.assets = assets
        assert ReplayManager.main(args) == Exit.SUCCESS
        assert "from_cmdline" in target.assets.assets
    assert target.launch.call_count == 1
    assert target.check_result.call_count == 1
    assert target.filtered_environ.call_count == 1
    assert serve_path.call_count == 1
    assert log_path.is_dir()
    assert any(log_path.glob("**/sample_asset"))


@mark.parametrize(
    "pernosco, rr, valgrind, no_harness",
    [
        # No debugger enabled and no harness
        (False, False, False, False),
        # No debugger enabled and with harness
        (False, False, False, True),
        # Pernosco enabled
        (True, False, False, False),
        # rr enabled
        (False, True, False, False),
        # Valgrind enabled
        (False, False, True, False),
    ],  # pylint: disable=invalid-name
)
def test_main_06(
    mocker, tmp_path, pernosco, rr, valgrind, no_harness
):  # pylint: disable=invalid-name
    """test ReplayManager.main() enable debuggers"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    # mock Sapphire.serve_path only
    mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path",
        autospec=True,
        return_value=(Served.ALL, ["test.html"]),  # passed to Target.check_result
    )
    # setup Target
    target = mocker.NonCallableMock(spec_set=Target, binary="bin", launch_timeout=30)
    target.check_result.return_value = Result.NONE
    load_target = mocker.patch(
        "grizzly.replay.replay.load_plugin",
        autospec=True,
        return_value=mocker.Mock(spec_set=Target, return_value=target),
    )
    # setup args
    (tmp_path / "test.html").touch()
    args = mocker.Mock(
        asset=[],
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=tmp_path / "test.html",
        launch_attempts=3,
        min_crashes=2,
        no_harness=no_harness,
        pernosco=pernosco,
        post_launch_delay=None,
        relaunch=1,
        repeat=1,
        rr=rr,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=valgrind,
    )
    # maximum one debugger allowed at a time
    assert sum((pernosco, rr, valgrind)) < 2, "test broken!"
    assert ReplayManager.main(args) == Exit.FAILURE
    assert target.check_result.call_count == 1
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1
    assert load_target.return_value.call_args[-1]["pernosco"] == pernosco
    assert load_target.return_value.call_args[-1]["rr"] == rr
    assert load_target.return_value.call_args[-1]["valgrind"] == valgrind


def test_main_07(mocker, tmp_path):
    """test ReplayManager.main() - report to FuzzManager"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    reporter = mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    # mock Sapphire.serve_path only
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path",
        autospec=True,
        return_value=(Served.ALL, ["test.html"]),  # passed to Target.check_result
    )
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    target = mocker.Mock(spec_set=Target, binary="bin", environ={}, launch_timeout=30)
    target.assets = mocker.Mock(spec_set=AssetManager)
    target.check_result.side_effect = (Result.FOUND,)
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    with TestCase("test.html", None, "adpt") as src:
        src.add_from_bytes(b"test", "test.html")
        src.dump(str(tmp_path / "testcase"), include_details=True)
    # setup args
    args = mocker.Mock(
        any_crash=False,
        asset=[],
        fuzzmanager=True,
        idle_delay=0,
        idle_threshold=0,
        ignore=[],
        input=tmp_path / "testcase",
        launch_attempts=1,
        logs=None,
        min_crashes=1,
        no_harness=False,
        pernosco=False,
        post_launch_delay=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
        tool=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Exit.SUCCESS
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 1
    assert target.check_result.call_count == 1
    assert serve_path.call_count == 1
    assert load_target.call_count == 1
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1
    assert target.assets.add.call_count == 0
    assert target.assets.is_empty.call_count == 1
    assert reporter.return_value.submit.call_count == 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import DEBUG, INFO

from pytest import mark

from .storage import TestCase
from .utils import (
    DEFAULT_TIME_LIMIT,
    TIMEOUT_DELAY,
    configure_logging,
    display_time_limits,
    grz_tmp,
    time_limits,
)


def test_grz_tmp_01(mocker, tmp_path):
    """test grz_tmp()"""
    fake_tmp = tmp_path / "grizzly"
    mocker.patch("grizzly.common.utils.GRZ_TMP", fake_tmp)
    # create temp path
    assert not fake_tmp.is_dir()
    path = grz_tmp()
    assert path == fake_tmp
    assert path.is_dir()
    # create temp path (exists)
    assert grz_tmp() == fake_tmp
    # create temp path with sub directory
    path = grz_tmp("test1", "test2")
    assert path == fake_tmp / "test1" / "test2"
    assert path.is_dir()


@mark.parametrize(
    "env, log_level",
    [
        # default log level
        ("0", INFO),
        # debug log level
        ("0", DEBUG),
        # enable debug log level via env
        ("1", INFO),
        # enable debug log level via env
        ("TRUE", INFO),
    ],
)
def test_configure_logging_01(mocker, env, log_level):
    """test configure_logging()"""
    config = mocker.patch("grizzly.common.utils.basicConfig", autospec=True)
    mocker.patch("grizzly.common.utils.getenv", autospec=True, return_value=env)
    configure_logging(log_level)
    assert config.call_count == 1
    if env != "0":
        assert config.call_args[-1]["level"] == DEBUG
    else:
        assert config.call_args[-1]["level"] == log_level


@mark.parametrize(
    "time_limit, timeout, test_durations, expected",
    [
        # use defaults
        (None, None, [None], (DEFAULT_TIME_LIMIT, DEFAULT_TIME_LIMIT + TIMEOUT_DELAY)),
        # use defaults instead of low test values
        (None, None, [1], (DEFAULT_TIME_LIMIT, DEFAULT_TIME_LIMIT + TIMEOUT_DELAY)),
        # use duration from test case
        (None, None, [99.1], (100, 100 + TIMEOUT_DELAY)),
        # multiple tests
        (None, None, [99.9, 10, 25], (100, 100 + TIMEOUT_DELAY)),
        # specify time limit
        (100, None, [0], (100, 100 + TIMEOUT_DELAY)),
        # specify timeout (> DEFAULT_TIME_LIMIT)
        (None, 100, [0], (DEFAULT_TIME_LIMIT, 100)),
        # specify timeout (< DEFAULT_TIME_LIMIT)
        (None, 10, [0], (10, 10)),
        # specify time limit and timeout
        (50, 100, [0], (50, 100)),
        # timeout disabled - use default time limit
        (None, 0, [None], (DEFAULT_TIME_LIMIT, 0)),
        # timeout disabled - with time limit
        (15, 0, [None], (15, 0)),
    ],
)
def test_time_limits_01(mocker, time_limit, timeout, test_durations, expected):
    """test time_limits()"""
    tests = [mocker.Mock(spec_set=TestCase, duration=d) for d in test_durations]
    assert time_limits(time_limit, timeout, tests=tests) == expected


@mark.parametrize(
    "time_limit, timeout, no_harness, msg",
    [
        # typical - harness
        (1, 2, False, "Using time limit: 1s, timeout: 2s"),
        # typical - without harness
        (1, 2, True, "Using timeout: 2s, harness: DISABLED"),
        # warn time limit and timeout match - harness
        (1, 1, False, "To avoid unnecessary relaunches set timeout > time limit"),
        # disabled timeout - harness
        (1, 0, False, "Using time limit: 1s, timeout: DISABLED"),
        # disable timeout - without harness
        (1, 0, True, "Using timeout: DISABLED, harness: DISABLED"),
    ],
)
def test_display_time_limits_01(caplog, time_limit, timeout, no_harness, msg):
    """test display_time_limits()"""
    display_time_limits(time_limit, timeout, no_harness)
    assert msg in caplog.text



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from grizzly.adapter import Adapter

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class NoOpAdapter(Adapter):
    """This is a simple adapter that has very little overhead. It can be used
    to help measure a baseline iteration rate for Grizzly.
    """

    NAME = "no-op"

    def setup(self, _input, _server_map):
        """Generate a static test case that calls `window.close()` when run.
        Normally this is done in generate() but since the test is static only
        do it once. Use the default harness to allow running multiple test cases
        in a row without closing the browser after each one.

        Args:
            _input (str): Unused.
            _server_map (sapphire.server_map.ServerMap): Unused.

        Returns:
            None
        """
        self.enable_harness()
        self.fuzz["test"] = (
            b"<!DOCTYPE html>\n"
            b"<html>\n"
            b"<head>\n"
            b"<script>window.close()</script>\n"
            b"</head>\n"
            b"</html>"
        )

    def generate(self, testcase, _server_map):
        """Since the test case has already been created just add the data to the
        TestCase.

        Also all TestCases require an entry point and the one expected by Grizzly
        is provided in `testcase.landing_page` so use it as the file name for
        the test.

        Args:
            testcase (grizzly.common.storage.TestCase): TestCase to be populated.
            _server_map (sapphire.server_map.ServerMap): Unused.

        Returns:
            None
        """
        testcase.add_from_bytes(self.fuzz["test"], testcase.landing_page)
